//** @file graphbuilder.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "graphbuilder.h"

#define Mb 1048576L

/**
 * This constructor allocates approximately (smem+jmem) Mb for the trie and
 * nmem Mb for the nodetable.
 */
GraphBuilder :: GraphBuilder(int smem, int jmem, int nmem, int lmem, bool sl) {

  trie = new Trie((smem * Mb), (jmem * Mb)/sizeof(SimpleHashPair<char*>)); 

  // this allows us to reference a WebNode by its URL
  long numels = (nmem * Mb)/sizeof(SimpleHashPair<WebNodePtr>);
  nodetable = new SimpleWebNodePtrHashTable(numels);

  // this is the web link graph
  graph = new WebLinkGraph();
  graph_is_docked = true;

  // these are used in constructing the nodes
  urlfilter = new URLFilter(true); // true means remove suffixes
  linkset = new RawLinkSet;

  curdoc = NULL;
  doc_url[0] = 0;
  doc_alias[0] = 0;
  curdoc_baseurl.Clear();

  // some statistics
  stats.heap_used_webnodes = 0;

  stats.cumulative_tolinks = 0;
  stats.cumulative_fromlinks = 0;
  stats.cumulative_dangling = 0;
  stats.cumulative_leaflinks = 0;

  stats.nodetable_insertions = 0;
  stats.nodetable_alias_insertions = 0;

  stats.lowid = kint32max;
  stats.highid = 0;
  stats.lowdate = kuint16max;
  stats.highdate = 0;

  flags.show_links = sl;
  flags.leaftable_memory = lmem;

}

GraphBuilder :: ~GraphBuilder() {
  // don't delete graph!
  delete linkset;
  delete nodetable;
  delete trie;
  if( leaftable ) { delete leaftable; }
}

void GraphBuilder :: StatisticsMem(ostream& o) {

  trie->Statistics(o);

  o << "\nTotal number of insertions in nodetable: " << stats.nodetable_insertions << endl;
  o << "Total number of aliases in nodetable:    " << stats.nodetable_alias_insertions << endl;
  o << "Heap needed for nodetable (bytes):       " << ((stats.nodetable_insertions + stats.nodetable_alias_insertions) * sizeof(SimpleHashPair<WebNodePtr>))<< endl;
  o << "Heap reserved for nodetable (bytes):     " << (nodetable->Size() * sizeof(SimpleHashPair<WebNodePtr>)) << endl;

  o << "\nMinimum size of each webnode (bytes):    " << sizeof(WebNode) << endl;
  o << "Heap needed for all webnodes (bytes):    " << stats.heap_used_webnodes << endl;
  o << "Unallocated webnodes (bytes):            " << WebNode::FreeBlocks() << endl;
  o << "Unallocated tolinks (bytes):             " << MemPool<LinkStruct>::FreeBlocks1() + MemPool<LinkStruct>::FreeBlocks2() << endl;
}

void GraphBuilder :: StatisticsGraph(ostream& o) {
  o << "Total number of nodes in web graph:      " << graph->size() << endl;
  o << "Lowest ID number in web graph:           " << stats.lowid << endl;
  o << "Highest ID number in web graph:          " << stats.highid << endl;
  o << "Lowest date in web graph:                " << stats.lowdate << endl;
  o << "Highest date in web graph:               " << stats.highdate << endl;

  o << "\nAverage number of tolinks/node:          " << static_cast<double>(stats.cumulative_tolinks)/graph->size() << endl;
  o << "Average number of dangling tolinks/node: " << static_cast<double>(stats.cumulative_dangling)/graph->size() << endl;
  o << "Average number of fromlinks/node:        " << static_cast<double>(stats.cumulative_fromlinks)/graph->size() << endl;
  o << "Average number of leaflinks/node:        " << static_cast<double>(stats.cumulative_leaflinks)/graph->size() << endl;
}

/// Creates a new WebNode on the heap and assigns anothernode as a handle
void GraphBuilder :: NodeInitialize(uint32 idno) {

  assert( !curdoc );

  curdoc = new WebNode(idno);
  doc_url[0] = 0;
  doc_alias[0] = 0;
  curdoc_baseurl.Clear();

  //another_merge = false;

  stats.lowid = min(stats.lowid, idno);
  stats.highid = max(stats.highid, idno);

}

/// Places the WebNode handled by anothernode into the graph and clears anothernode.
/**
 * The WebNode is pushed at the front of the graph. Since node id's are given sequentially
 * in increasing order, the graph will contain nodes with decreasing id sequence. This ordering
 * should not be tampered with, as it is used by Talker().
 */
void GraphBuilder :: NodeLaunch() {

   assert( curdoc );

   stats.lowdate = (curdoc->Date() > 0) ? 
     min(stats.lowdate, curdoc->Date()) : stats.lowdate;
   stats.highdate = max(stats.highdate, curdoc->Date());

   graph->push_front(curdoc);
   linkset->clear();

   curdoc = NULL;
   doc_url[0] = 0;
   doc_alias[0] = 0;
   curdoc_baseurl.Clear();
   //another_merge = false;
}

/// Sets anothernode's date.
void GraphBuilder :: NodeSetDate(unsigned short adate) {
  curdoc->SetDate(adate);
}

/// Inserts the anchor links contained in linkset into anothernode's fromlinks array.
/**
 * Note that at this stage, all the links are pointer differences into the trie.
 */
void GraphBuilder :: NodeInsertLinks() {
  curdoc->InsertRawLinks(linkset);
}

/// Copies the current document's URL into another_url, and places it into the trie.
/**
 * The copy is necessary so that later calls to FormatURL() can use another_url
 * to complete anchor link URLs, whenever those are incomplete.
 */
void GraphBuilder :: NodeSetURL(const char *docurl, const char *aliasurl) {

  docurl__ = docurl;
  aliasurl__ = aliasurl;

  assert( !strlen(doc_url) );

  // format the document's url for insertion in the trie
  // formatURL returns a string whose length is guaranteed less than STRINGBUF_LEN2
  ContentType contype;
  strcpy(doc_url, urlfilter->FormatURL(docurl, strlen(docurl), NULL, &contype));

  // insert the url both in trie and the nodetable
  // in case of duplicates, exchange the curdoc pointer with 
  // a pointer to the already existing WebNode
  {
    ptrdiff_t key = trie->InsertURL(urlfilter->CompressURL(urlfilter->DeindexURL(doc_url)));

    WebNodePtr w = nodetable->Find(key);

    if( !w ) {

      assert(curdoc);

      nodetable->Insert(key, curdoc);

      stats.nodetable_insertions++;
    } else {  // this node was already inserted
      cerr << "warning: aliased or duplicate node [" << doc_url << "]" << endl;
      // this is called at the beginning when the link pointers haven't
      // been allocated yet
      delete curdoc; // rid of current node
      curdoc = w; // replace with existing one with same URL
      //another_merge = true; // make sure new links are added to old ones
    }
  }

  // compute the current document's base URL address components
  // this is used when formatting anchor links.
  urlfilter->ParseURL(doc_url, curdoc_baseurl.scheme, curdoc_baseurl.netloc,
		      curdoc_baseurl.query, curdoc_baseurl.params, curdoc_baseurl.path);

  // this will make aliasurl point to curdoc also
  if( aliasurl ) {
    strcpy(doc_alias, urlfilter->FormatURL(aliasurl, strlen(aliasurl), NULL, &contype));

    ptrdiff_t key = trie->InsertURL(urlfilter->CompressURL(urlfilter->DeindexURL(doc_alias)));

    WebNodePtr w = nodetable->Find(key);
    if( !w ) {
      assert(curdoc);
      nodetable->Insert(key, curdoc);
      stats.nodetable_alias_insertions++;
    } else {
      // do nothing
      if(w != curdoc) {
	cerr << "warning: alias points to a different node. Dataset inconsistent? [" << doc_alias << "]" << endl;
	// policy: we ignore the error
      };
    }
  }

  if( flags.show_links ) {
    cout << doc_url << endl;
  }
}

const char * GraphBuilder :: NodeGetURL_() {
  return docurl__;
}

const char * GraphBuilder :: NodeGetAlias_() {
  return aliasurl__;
}

const char * GraphBuilder :: NodeGetURL() {
  return doc_url;
}

const char * GraphBuilder :: NodeGetAlias() {
  return doc_alias;
}

const uint32 GraphBuilder :: NodeGetID() {
  return curdoc->ID();
}


const uint16 GraphBuilder :: NodeGetDate() {
  return curdoc->Date();
}

/// Here we walk through all the nodes and change the
/// links so that they no longer require the trie.
WebLinkGraph* GraphBuilder :: UndockWebGraph() {
  cerr << "info: now undocking webgraph" << endl;
  if( graph_is_docked ) {
    // zeroth pass. Normalize tolinks and calculate array sizes for fromlinks
    for(WebLinkGraph::iterator i = graph->begin(); i != graph->end(); i++) {
      assert(*i);
      // convert ptrdiff_t to webnode* for all valid links
      // also initializes the valid link count
      (*i)->NormalizeRawLinks(nodetable);
     }

    // first pass. Calculate sizes of fromlinks arrays
     for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
       assert(*i);
      for(int k = 0; k < (*i)->NumberOfValidToLinks(); k++) {
	assert((*i)->ValidToLink(k));
	((*i)->ValidToLink(k))->IncrementNumberOfFromLinks();
      }
    }

    // second pass. Insert fromlinks and calculate statistics
    for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
      assert(*i);
      for(int k = 0; k < (*i)->NumberOfValidToLinks(); k++) {
	assert((*i)->ValidToLink(k));
	((*i)->ValidToLink(k))->AppendFromLink(*i);
      }
      stats.heap_used_webnodes += (*i)->RealSize();
      stats.cumulative_dangling += (*i)->NumberOfDanglingToLinks();
      stats.cumulative_fromlinks += (*i)->NumberOfValidFromLinks();
      stats.cumulative_leaflinks += (*i)->NumberOfLeafLinks();
      stats.cumulative_tolinks += (*i)->NumberOfValidToLinks();
    }

#ifndef NDEBUG
    // third pass. check that fromlinks work properly
    cerr << "debug: check that fromlinks work properly" << endl;
    for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
      for(int k = 0; k < (*i)->NumberOfValidFromLinks(); k++) {
	assert((*i)->ValidFromLink(k));
      }
    }
#endif
    graph_is_docked = false;
  }
  cerr << "info: webgraph is now undocked" << endl;
  return graph;
}

void GraphBuilder :: UpdateLeafLinks() {
  assert(leaftable);
  stats.cumulative_dangling = 0;
  stats.cumulative_leaflinks = 0;
  for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
    assert(*i);
    (*i)->UpdateLeafLinks(leaftable);
    stats.cumulative_leaflinks += (*i)->NumberOfLeafLinks();
    stats.cumulative_dangling += (*i)->NumberOfDanglingToLinks();
  }

#ifndef NDEBUG
    // third pass. check that leaflinks work properly
    cerr << "debug: check that leaflinks work properly" << endl;
    for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
      for(int k = 0; k < (*i)->NumberOfLeafLinks(); k++) {
	assert((*i)->ValidLeafLink(k));
	assert((*i)->ValidLeafLink(k)->OccupationCount() == 0);
      }
    }
#endif

}

/// finds an existing WebNode from a URL. Returns NULL if not found.
WebNode* GraphBuilder :: FindWebNode(const char *url) {
  ptrdiff_t key = trie->FindURL(urlfilter->CompressURL(urlfilter->DeindexURL(url)));
  if( key > -1 ) {
    WebNodePtr w = nodetable->Find(key);
    if( w ) {
      return w;
    }
  }
  return NULL;
}

/// finds an existing LeafNode key into the leaftable from a URL. Returns -1 if not found.
const ptrdiff_t GraphBuilder :: FindLeafNodeKey(const char *url) {
  return trie->FindURL(urlfilter->CompressURL(urlfilter->DeindexURL(url)));
}

void GraphBuilder :: SetupLeafTable() {
  if( !leaftable ) {
    leaftable = new SimpleLeafNodePtrHashTable((flags.leaftable_memory * Mb)/sizeof(SimpleHashPair<LeafNodePtr>));
  }
  leaftable->Clear();
}

void GraphBuilder :: AddLeaf(const ptrdiff_t key, LeafNodePtr leaf) {
  assert(leaftable);
  if( !leaftable->Find(key) ) {
    leaftable->Insert(key, leaf);
  } else {
    cerr << "warning: duplicate leaf ignored" << endl;
  }

  stats.lowdate = (leaf->Date() > 0) ? 
    min(stats.lowdate, leaf->Date()) : stats.lowdate;
  stats.highdate = max(stats.highdate, leaf->Date());
}

void GraphBuilder :: TrieInsertLinkURL(const char *url) {
  ptrdiff_t key = trie->InsertURL(urlfilter->CompressURL(urlfilter->DeindexURL(url)));
  assert(*(trie->bigs + key) == 0);
  linkset->insert(key);
}

