/** @file graphbuilder.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _GRAPH_BUILDER_H_
#define _GRAPH_BUILDER_H_
#include "document.h"
#include "webnode.h"
#include "leafnode.h"
#include "leafnode.h"
#include "simplehash.h"
#include "linkgraph.h"
#include "trie.h"
#include "urlfilter.h"

#define SLASHBIT 0x80

/// Hashtable used to look up WebNode objects from its URL in the trie.
typedef SimpleHashTable<WebNodePtr> SimpleWebNodePtrHashTable;

/// Builds the web link graph as an object of type WebLinkGraph.
/**
 * Every WebNode is a separate document, and GraphBuilder handles
 * the connection of fromlinks and tolinks, and the discarding of
 * dangling links. 
 *
 * To do this, GraphBuilder must construct a trie containing all 
 * known document URLs. Once the WebLinkGraph is built, the trie
 * is "undocked" from the list. This allows the (substantial) memory
 * taken by the URL strings to be regained, at the cost of no longer 
 * being able to identity a WebNode by its document URL.
 */
class GraphBuilder {
 public:

  GraphBuilder(int smem, int jmem, int nmem, int lmem, bool sl);
  ~GraphBuilder();

  void NodeInitialize(uint32 idno);
  void NodeSetURL(const char *docurl, const char *aliasurl);
  const char* NodeGetURL();
  const char* NodeGetAlias();
  const char* NodeGetURL_();
  const char* NodeGetAlias_();
  const uint32 NodeGetID();
  const uint16 NodeGetDate();
  void NodeSetDate(unsigned short aDate);
  void NodeInsertLinks();
  void NodeLaunch();
  URLComponents* NodeGetURLParts() 
    { return &curdoc_baseurl; }
  
  WebLinkGraph* UndockWebGraph();

  const char* FormatURL(const char *anurl, int anurl_len, ContentType *t) 
    { return urlfilter->FormatURL(anurl, anurl_len, &curdoc_baseurl, t); }

  void TrieInsertLinkURL(const char *url);

  WebNodePtr FindWebNode(const char *url);

  void SetupLeafTable();
  void AddLeaf(ptrdiff_t key, LeafNodePtr leaf);
  const ptrdiff_t FindLeafNodeKey(const char *url);
  void UpdateLeafLinks();

  void StatisticsMem(ostream& o);
  void StatisticsGraph(ostream& o);

  uint32 LowestID()
    { return stats.lowid; }
  uint32 HighestID()
    { return stats.highid; }

  uint16 LowestDate()
    { return stats.lowdate; }
  uint16 HighestDate()
    { return stats.highdate; }

  struct {
    bool show_links;
    int leaftable_memory;
  } flags;

 private:

  WebNode* curdoc;

  URLComponents curdoc_baseurl;
  char doc_url[STRINGBUF_LEN2+1];
  char doc_alias[STRINGBUF_LEN2+1];

  const char *docurl__;
  const char *aliasurl__;
  //bool another_merge; // for when we insert duplicate nodes 
  // store its decomposition for easy access

  WebLinkGraph *graph;
  bool graph_is_docked;

  Trie *trie;
  SimpleWebNodePtrHashTable *nodetable;
  SimpleLeafNodePtrHashTable *leaftable;
  URLFilter *urlfilter;

  RawLinkSet *linkset; 

  struct {
    uint32 heap_used_webnodes;
    uint32 cumulative_tolinks;
    uint32 cumulative_fromlinks;
    uint32 cumulative_leaflinks;
    uint32 cumulative_dangling;
    uint32 nodetable_insertions;
    uint32 nodetable_alias_insertions;
    uint32 lowid;
    uint32 highid;
    uint16 lowdate;
    uint16 highdate;
  } stats;

};

#endif
