/** @file jack.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include <string>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sstream>
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>

static char defaultfifo[] = "";
static char defaulttmpdir[] = "/tmp/";

/// Opens both input and output FIFOs and sends commands to ripper
int main(int argc, char** argv) {

  char *name = defaultfifo;
  char *tempdir = defaulttmpdir;

  argv++;
  for (int i = 1; i < argc; ++i, ++argv) {
    if ((*argv)[0] == '-' && (*argv)[1] != '\0') {  // option
      if (!strcmp(*argv, "--name")) {
        ++i;
        ++argv;
        if (i >= argc) {
          cerr << "usage: jack [--name monty]\n";
	  exit(1);
        }
        name = *argv;
      } else if (!strcmp(*argv, "--temp_dir")) {
        ++i;
        ++argv;
        if (i >= argc) {
          cerr << "usage: jack [--temp_dir /tmp/]\n";
	  exit(1);
        }
        tempdir = *argv;
      }
    }
  }
  ostringstream inputfifo;
  ostringstream outputfifo;

  char *command;
  char prompt[30];

  int ifd, ofd;
  fd_set iset, oset;

  // this is to liven up the prompt
  time_t tstart, tstop;

  inputfifo << tempdir << "ripper";
  if( *name ) {
    inputfifo << "." << name;
  }
  inputfifo << ".input";

  outputfifo << tempdir << "ripper";
  if( *name ) {
    outputfifo << "." << name;
  }
  outputfifo << ".output";

  // opening the FIFOs will block unless
  // the other end is already opened.
  // To prevent deadlock with ripper,
  // we mut open in nonblocking mode here
  if((ifd = open(outputfifo.str().c_str(), O_RDONLY|O_NONBLOCK)) < 0) {
    cerr << "error: FIFO cannot be opened [" << inputfifo.str() << "]" << endl;
    cerr << "is ripper listening?" << endl;
    exit(1);
  }
  if((ofd = open(inputfifo.str().c_str(), O_WRONLY|O_NONBLOCK|O_APPEND)) < 0) {
    cerr << "error: FIFO cannot be opened [" << outputfifo.str() << "]" << endl;
    cerr << "is ripper listening?" << endl;
    exit(1);
  }

  cout << "Welcome to jack. Command tokens are separated by spaces."
    "\nType help for a list of commands." << endl;

  FD_ZERO(&iset); FD_SET(ifd, &iset);
  FD_ZERO(&oset); FD_SET(ofd, &oset);

  // initialize readline

  char c;
  int size;
  bool quit = false;
  tstart = 0;
  while(!quit) {

    int retval = select(ifd+1, &iset, NULL, NULL, NULL); 
    if( retval > 0 ) {
      // data available
      if( (size = read(ifd, &c, 1)) >= 0) {
	if( c == '\r' ) { // end of communication marker
	  time(&tstop);
	  if( tstart > 0 ) {
	    sprintf(prompt, "\ncomputation took %g seconds\n> ", 
		    difftime(tstop, tstart));
	  } else {
	    strcpy(prompt, "\n> ");
	  }
	  command = readline(prompt);
	  if( command && *command ) {
	    int k = 0;
	    int n = strlen(command);
	    do {
	      retval = select(ofd+1, NULL, &oset, NULL, NULL);
	      if( retval ) {
		size = write(ofd, command + k, strlen(command + k));
		k += size;
	      }
	    } while( k < n );
	    if( strncmp(command,"quit",4) == 0) {
	      quit = true;
	    }
	    add_history(command);
	    free(command);
	    command = (char *)NULL;
	  }
	  size = write(ofd, "\n", 1);
	  time(&tstart);
	} else {
	  cout << c;
	}
      } // else bad read
    }
  }

  // if we exit before ripper, the fifos don't get unlinked
  struct stat buf1, buf2;
  do {
    sleep(1);
  } while(fstat(ifd, &buf1) || fstat(ofd, &buf2));

  return 0;
}

