/** @file linkgraph.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "webnode.h"
#include "linkgraph.h"
#include <math.h>

WebLinkGraph :: WebLinkGraph(): WebNodeList() {
  for(int k =0; k < TAG_NUMBER_OF_BITS; k++) {
    fromsetsize[k] = 0;
  }
}

/// Prints to a stream a textual description of the graph structure
void WebLinkGraph :: PrintWebGraph(ostream& o) {
  o << "# node | date | num_dangling | num_fromlinks | num_tolinks | tolinks_list..." << endl;
  for(WebNodeList::iterator i = begin(); i != end(); i++ ) {
    o << (*i)->ID()
      << " " << (*i)->Date()
      << " " << (int)(*i)->NumberOfDanglingToLinks()
      << " " << (*i)->NumberOfValidFromLinks()
      << " " << (*i)->NumberOfValidToLinks();
    int n = (*i)->NumberOfValidToLinks();
    for(int k = 0; k < n; k++) {
      o << " " << (*i)->ValidToLink(k)->ID();
    }
    for(int k = 0; k < (*i)->NumberOfLeafLinks(); k++) {
      o << " " << (*i)->ValidLeafLink(k)->ID();
    }
    o << endl;
  }
}

/// Clears all tagged WebNode objects
void WebLinkGraph :: ClearTags() {
  for(WebNodeList::iterator i = begin(); i != end(); i++) {
    (*i)->ClearTag();
  }
}

/// Builds FromSets for a given set of tagged WebNode objects.
/// Note parameter k is currently not used.
void WebLinkGraph :: BuildFromSets(int k) {
  // k not actually used
  assert( k <= TAG_NUMBER_OF_BITS );
  assert( k >= 0 );
  for(int t = 0; t < (TAG_NUMBER_OF_BITS - 1); t++) {
    cerr << "info: tagging fromset # " << (t+1) << endl;
    for(WebNodeList::iterator i = begin(); i != end(); i++) {
      assert(*i);
      WebNodePtr w = (*i);
      assert(w);
      if( w->Tagged(t) ) {
	for(int j = 0; j < w->NumberOfValidFromLinks(); j++) {
	  WebNodePtr v = w->ValidFromLink(j);
	  assert(v);
	  // assert(v != w); NOT true, since redirects can introduce autolinks
	  v->SetTag(t+1);
	  assert( v->Tagged(t+1) );
	}
      }
    }
  }
}

/// Calculates the size of each FromSet.
/// This calculation is saved in the array fromsetsize[]
void WebLinkGraph :: MeasureFromSets() {
  // clear counts
  for(int j = 0; j < TAG_NUMBER_OF_BITS; j++) {
    fromsetsize[j] = 0;
  }
  // iterate through the graph
  for(WebNodeList::iterator i = begin(); i != end(); i++) {
    assert(*i);
    for(int j = 0; j < TAG_NUMBER_OF_BITS; j++) {
      if( (*i)->Tagged(j) ) {
	fromsetsize[j]++;
      }
    }
  }
}

/// Lists the fromsetsize[] values to a stream.
void WebLinkGraph :: StatisticsFromSets(ostream& o) {
  o << "\nNumber of tagged webnodes:  " << fromsetsize[0] << endl;
  if( fromsetsize[0] ) {
    for(int i = 1; i < TAG_NUMBER_OF_BITS; i++) {
      o << "Number of webnodes at distance " << i << " from tagged set:  " << fromsetsize[i] << endl;
    }
  }
}
