/** @file mempool.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _MEM_POOL_H_
#define _MEM_POOL_H_

#include "basictypes.h"
#include <cassert>
#include <stdexcept>
#include <stdlib.h>
#include <new>
#include <iostream>

/// Wraps around a class T and redefines new() and delete()
/// member functions.
/** Allocates chunks of BLOCK_SIZE objects at a time, then
 *  gives them out and recycles them. 
 */
template<class T>
class MemoryPooled
{
 public:
  MemoryPooled() {} // default constructor does nothing
  ~MemoryPooled() {} // we'll never call this :-)
 
  static void* operator new(size_t size); //size_t size);
  static void operator delete(void *dead, size_t size);

  static int FreeBlocks();

 protected:
  union {
    T data;
    MemoryPooled<T> *next;
  };

 private:

  static const int BLOCK_SIZE;
  static MemoryPooled<T>* allocate_block();
  static MemoryPooled<T> *freelist;
};

/// Allocate a single object
template<class T>
void* MemoryPooled<T> :: operator new(size_t size) {

  if( size != sizeof(MemoryPooled<T>)) { 
    cerr << "mempool: allocated object of nonstandard size" << endl;
    return ::operator new(size); 
  }
  MemoryPooled<T> *p = freelist;
  if(p) { 
      freelist = p->next;
  } else {
    assert(BLOCK_SIZE > 1);
    p = allocate_block();
    freelist = p->next;
    assert(freelist);
  }
  assert(p);
  return p;
}

/// Utility function
template<class T>
MemoryPooled<T>* MemoryPooled<T> :: allocate_block() {
  MemoryPooled<T> *newBlock = 
    static_cast<MemoryPooled<T>*>(::operator new(BLOCK_SIZE * 
						 sizeof(MemoryPooled<T>)));
  // this memset is so it shows up in top
  memset(newBlock,0,BLOCK_SIZE * sizeof(MemoryPooled<T>));
  for(int i = 0; i < BLOCK_SIZE - 1; i++) {
    newBlock[i].next = &newBlock[i+1];
  }
  newBlock[BLOCK_SIZE - 1].next = 0;
  return newBlock;
}

/// Recycle a single object
template<class T>
void MemoryPooled<T> :: operator delete(void *dead, size_t size) {

  if( dead == 0 ) { 
    return; 
  }
  if( size != sizeof(MemoryPooled<T>) ) {
    cerr << "warning: overloaded delete called global delete" << endl;
   :: operator delete(dead);
    return;
  }
  MemoryPooled<T> *shell = static_cast<MemoryPooled<T>*>(dead);
  shell->next = freelist;
  freelist = shell;
}


template<class T>
int MemoryPooled<T> :: FreeBlocks() {
  //  cout << "begin freeblocks1" << endl;
  MemoryPooled<T> *p = freelist;
  int fsize = 0;
  while( p ) {
    fsize += sizeof(MemoryPooled<T>);
    p = p->next;
  }
  return fsize;
}

//==========================================//

/// A memory management wrapper around a class S.
/**
 * The wrapper adds two admin data members to S, as a union.
 * These members allow the class MemPool<S> to keep track
 * of free allocated memory blocks in a linked list.
 * 
 * If you need a large number of small (one to two element)
 * arrays of type S, wrap S as a MemPoolObject<S> and allocate
 * memory using a MemPool<S> object. Not useful if sizeof(S) < 4.
 */
template<class S>
union MemPoolObject {
  S data;
  MemPoolObject<S>* admin_next;
  int admin_length;
};

/// A memory managed class suitable for many small arrays of type S.
/**
 * This memory pool class handles memory management for large
 * numbers of small arrays whose elements are of type S.
 * The normal new[] and delete[] functions for S objects 
 * introduce much
 * overhead, so they aren't suitable for the "fromlinks" and "tolinks"
 * Link arrays associated with each WebNode. Instead, MemPool<S> 
 * provides Allocate() and Deallocate() member functions to replace
 * new[] and delete[]. 
 */
template<class S>
class MemPool
{

 public:
  MemPool() {} // default constructor does nothing
  ~MemPool() {} // we'll never call this :-)
 
  MemPoolObject<S>* Allocate(int numels) throw (range_error);
  void Deallocate(MemPoolObject<S>* dead, int numels);

  static int FreeBlocks1();
  static int FreeBlocks2();
 private:

  static const int BLOCK_SIZE;
  static MemPoolObject<S>* allocate_block1();
  static MemPoolObject<S>* allocate_block2();

  static MemPoolObject<S> *mempool1; ///< mempool1 contains only single chunks
  static MemPoolObject<S> *mempool2; ///< mempool2 contains blocks of chunks suitable for arrays
};

template<class S>
MemPoolObject<S>* MemPool<S> :: Allocate(int numels) throw (range_error) {
  if( (numels <= 0) ) {
    cerr << "warning: overloaded new[] called global new" << endl;;
    return static_cast<MemPoolObject<S>*>(::operator new[](numels * sizeof(MemPoolObject<S>)));
  } else if( numels == 1 ) {
    // too small for freelist2
    MemPoolObject<S> *p = mempool1;
    if(p) { 
      mempool1 = p[0].admin_next; 
    } else {
      p = allocate_block1();
      assert(p);
      mempool1 = p[0].admin_next;
    }
    return p;
  }

  // each request now is for at least two elements
  // find the first block big enough
  MemPoolObject<S> *p = mempool2;
  if(p) {
    // find first chunk that's big enough for us
    MemPoolObject<S> *prev = NULL;
    while( p && (p[1].admin_length < numels) ) {
      prev = p;
      p = p[0].admin_next;
    }
    if( p ) { // found one
      // remove p block from the freelist2
      if( prev ) {
	prev[0].admin_next = p[0].admin_next;
      } else {
	mempool2 = p[0].admin_next;
      }

      // prepend the rest of the block minus numels
      // to freelist2, or freelist1 if it's too small
      int remaining = p[1].admin_length - numels;
      if( remaining == 1 ) {
	//cout << "mucking with freelist1" << endl;
	p[numels].admin_next = mempool1;
	mempool1 = &p[numels];
      } else if( remaining > 1 ) {
	p[numels].admin_next = mempool2;
	p[numels + 1].admin_length = remaining;
	mempool2 = &p[numels];
      }
      // return the p block
      return p;
    }
  }

  // if we're here, then we must allocate a new block
  assert(BLOCK_SIZE > 1); // don't want no riffraff here
  if(numels > BLOCK_SIZE) {
    cerr << "BLOCK_SIZE is too small" << endl;
    throw range_error("");
  } else {
    p = allocate_block2();
    // attach remaining block to the freelist
    int remaining = BLOCK_SIZE - numels;
    if( remaining == 1 ) { 
      p[numels].admin_next = mempool1;
      mempool1 = &p[numels];
    } else if( remaining > 1 ) {
	p[numels].admin_next = mempool2;
	p[numels + 1].admin_length = remaining;
	mempool2 = &p[numels];
    }
    return p;
  }
}

template<class S>
MemPoolObject<S>* MemPool<S> :: allocate_block1() {
  MemPoolObject<S> *newBlock = 
    static_cast<MemPoolObject<S>*>(::operator new(BLOCK_SIZE * sizeof(MemPoolObject<S>)));
  // this memset is so it shows up in top
  memset(newBlock,0,BLOCK_SIZE * sizeof(MemPoolObject<S>));
  for(int i = 0; i < BLOCK_SIZE - 1; i++) {
    newBlock[i].admin_next = &newBlock[i+1];
  }
  newBlock[BLOCK_SIZE - 1].admin_next = 0;
  return newBlock;
}

template<class S>
MemPoolObject<S>* MemPool<S> :: allocate_block2() {
  MemPoolObject<S> *newBlock = 
    static_cast<MemPoolObject<S>*>(::operator new(BLOCK_SIZE * sizeof(MemPoolObject<S>)));
  memset(newBlock,0,BLOCK_SIZE * sizeof(MemPoolObject<S>));
  newBlock[0].admin_next = 0;
  newBlock[1].admin_length = BLOCK_SIZE;
  return newBlock;
}

template<class S>
int MemPool<S> :: FreeBlocks1() {
  MemPoolObject<S> *p = mempool1;
  int fsize = 0;
  while( p ) {
    fsize += sizeof(MemPoolObject<S>);
    p = p[0].admin_next;
  }
  return fsize;
}

template<class S>
int MemPool<S> :: FreeBlocks2() {
  MemPoolObject<S> *p = mempool2;
  int fsize = 0;
  while( p ) {
    fsize += p[1].admin_length * sizeof(MemPoolObject<S>);
    p = p[0].admin_next;
  }
  return fsize;
}

template<class S>
void MemPool<S> :: Deallocate(MemPoolObject<S> *dead, int numels) {
  if( numels == 1 ) {
    dead[0].admin_next = mempool1;
    mempool1 = dead;
  } else if( numels > 1 ) {
    // order is important
    dead[1].admin_length = numels;
    dead[0].admin_next = mempool2;
    mempool2 = dead;
  }
}

#endif
