/** @file parsehandler-graph.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "handler-parser.h"
#include "graphbuilder.h"
#include "content-type.h"

/// A custom case insensitive strncmp
extern int strincmp(const char *s1, const char *s2, int s2len);

/// Reads anchor links and other document information for use by GraphBuilder.
/**
 * GraphParseHandler is called for each document being processed, and 
 * interfaces with GraphBuilder, telling it the current document's 
 * URL and anchor links, as well as the document's date.
 *
 * Not all anchor links are passed on to GraphBuilder, however. GraphParseHandler
 * doesn't pass on autolinks, or links to non-html documents.
 */
class GraphParseHandler : public ParseHandler {
 public:

  GraphParseHandler(GraphBuilder *graphbuilder);

  virtual void NewDocument(const Document* doc);
  virtual void AddHeader(const char* key, int keylen,
			 const char* value, int valuelen);
  virtual void AddAnchor(const char* href, int hreflen);

private:
  GraphBuilder * gb;
};

ParseHandler* MakeGraphHandler(GraphBuilder *graphbuilder) {
  return new GraphParseHandler(graphbuilder);
}

GraphParseHandler :: GraphParseHandler(GraphBuilder *graphbuilder) {
  assert( graphbuilder );
  gb = graphbuilder;
}

void GraphParseHandler :: NewDocument(const Document* doc) {

  // we use doc-url() but prefer doc->url_after_redirects()

  if( strlen(doc->url_after_redirects()) > 0 ) {
    gb->NodeSetURL(doc->url_after_redirects(), doc->url());
    //    cerr << "info: detected a redirected document url" << endl;
  } else {
    gb->NodeSetURL(doc->url(), NULL);
  }
}

/// Reads the date and calls gb->NodeSetDate()
void GraphParseHandler :: AddHeader(const char* key, int keylen,
				    const char* value, int valuelen) {
  // if we're parsing the date, convert
  // to number of days (julian date) and place into current WebNode
  if( (strncmp(key,"Date",4) == 0) ||
      (strncmp(key,"Last-Modified",13) == 0) ||
      (strncmp(key,"X-Google-Crawl-Date",19) == 0) ) {

    // assume date format Fri, 07 Sep 2001 04:55:16 GMT
    // also accept Wednesday, 14-Nov-01 00:53:22 GMT
    // anything else is deemed a bad date
    assert( valuelen < 50 );
    char mydate[50];
    char * p;
    int d=0, m=0, y=0;
    bool badformat = false;
    int64 jd;
    char * daytable = "janfebmaraprmayjunjulaugsepoctnovdec";

    // stupid strtok modifies string
    strncpy(mydate,value,valuelen);
    mydate[valuelen] = 0;

    // actually, if we blank out the time stamp, then we can parse more date formats
    for( p = mydate; *p; p++) {
      if( isdigit(p[0]) && isdigit(p[1]) && (p[2] == ':') && 
	  isdigit(p[3]) && isdigit(p[4]) && (p[5] == ':') &&
	  isdigit(p[6]) && isdigit(p[7]) && (!isdigit(p[8])) ) {
	memset(p, ' ', 8);
	break;
      }
    } 

    // now start tokenizing
    if( !strtok(mydate, " -") ) { // ignore day name
      badformat = true;
    }

    for(int daymon=0; daymon < 2; daymon++) {
      p =  strtok(NULL, " -"); // day or month
      if( p && isdigit(p[0]) ) { 
	d = atoi(p);
	if( (d > 31) || (d < 1) ) {
	  badformat = true;
	}
      } else if( p && isalnum(p[0]) ) {
	m = 0;
	for(int k = 0; k < 12; k++) {
	  int q = 3*k;
	  if( (daytable[q] == tolower(p[0])) &&
	      (daytable[q+1] == tolower(p[1])) &&
	      (daytable[q+2] == tolower(p[2])) ) {
	    m = k+1;
	    break;
	  }
	}
	if( m == 0 ) {
	  badformat = true;
	}
      } else {
	badformat = true;
      }
    }

    p = strtok(NULL, " -"); // year
    if( p ) {
      y = atoi(p);
      if( y < 100 ) {
	y = (y < 50) ? (y + 2000) : (y + 1900);
      } else if( y < 1994) {
	if( y == 100 ) {
	  y = 2000;
	} else if( y == 101 ) {
	  y = 2001;
	} else {
	  badformat = true;
	}
      } else if( y > 2010) {
	badformat = true;
      }
    } else {
      badformat = true;
    }

    if( !badformat ) {
      // convert to julian day number, since 15 Oct 1582
      jd = ( 1461 * ( y + 4800 + ( m - 14 ) / 12 ) ) / 4 +
	( 367 * ( m - 2 - 12 * ( ( m - 14 ) / 12 ) ) ) / 12 -
	( 3 * ( ( y + 4900 + ( m - 14 ) / 12 ) / 100 ) ) / 4 +
	d + 2367925;
      jd -= 4845000; // convert to int - offset arbitrary
      // DEBUG cout << d << " " << m << " " << y << " " << jd << endl;
    } else {
      jd = 0;
      strncpy(mydate,value,valuelen); // trailing null is still there
      cerr << "warning: cannot parse date, setting to zero [" << mydate << "]" << endl;
    }

    // now inser the date into the new node
    assert( jd <= kuint16max );
    gb->NodeSetDate((uint16)jd);
  }
}

void GraphParseHandler :: AddAnchor(const char* href, int hreflen) {
  // here we try to classify the type of anchor link, to see if it's
  // worth adding to the web graph. We *don't* convert %-encoded characters,
  // so obfuscated urlst are not treated properly - tough. Also, we don't
  // take particularly good care about case handling, which depends upon
  // the server's filesystem conventions.
  if( hreflen <= 0 ) { // wtf? discard
    return;
  } else if( href[0] == '#' ) { // link back to itself, discard
    return;
  } else if( (hreflen > 6) && (strncasecmp(href, "ftp://", 6) == 0) ) { 
    // most likely data file, discard
    return;
  } else if( (hreflen > 9) && (strncasecmp(href, "gopher://", 9) == 0) ) { 
    // foreign protocol, no hyperlinks
    return;
  } else if( (hreflen > 7) && (strncasecmp(href, "mailto:", 7) == 0) ) { 
    // no hyperlinks
    return;
  } else if( (hreflen > 5) && (strncasecmp(href, "nntp:", 5) == 0) ) { 
    // foreign protocol
    return;
  } else if( (hreflen > 5) && (strncasecmp(href, "news:", 5) == 0) ) { 
    // like nntp
    return;
  } else if( (hreflen > 7) && (strncasecmp(href, "telnet:", 7) == 0) ) { 
    // foreign protocol, no hyperlinks
    return;
  } else if( (hreflen > 7) && (strncasecmp(href, "tn3270:", 7) == 0) ) { 
    // like telnet
    return;
  } else if( (hreflen > 5) && (strncasecmp(href, "wais:", 5) == 0) ) { 
    // foreign
    return;
  } else if( (hreflen > 5) && (strncasecmp(href, "file:", 5) == 0) ) { 
    // this is debatable whether to accept, since some people
    // link html files this way
    char *p = strstr(href,"htm");
    if( p && ((p - href) < hreflen) ) { 
      // seems to contain an (s)htm(l) file
    } else { // discard
      return;
    }
  } else if( (hreflen > 11) && (strncasecmp(href, "javascript:", 11) == 0) ) { 
    // junk, discard - Die X10!
    return;
  } else if( (hreflen > 9) && (strncasecmp(href, "prospero:", 11) == 0) ) { 
    // foreign, discard
    return;
  }

  ContentType ctype;
  const char *url = gb->FormatURL(href, hreflen, &ctype);

  // throw away images and postscript documents, but keep GOOGLE_OTHER,
  // which means we didn't know how to classify the link
  switch (ctype) {
  case CONTENT_APPLICATION_POSTSCRIPT:
  case CONTENT_APPLICATION_PDF:
  case CONTENT_APPLICATION_MSWORD:
  case CONTENT_TEXT_RTF:
  case CONTENT_APPLICATION_MS_POWERPOINT:
  case CONTENT_APPLICATION_XGZIP:
  case CONTENT_IMAGE:
  case CONTENT_TEXT_PLAIN:
  case CONTENT_AUDIO_MP3:
    return;
    break;
  case CONTENT_GOOGLE_OTHER:
  case CONTENT_TEXT_HTML:
  default:
    break;
    }

  // don't allow a link to the current document
  if( strcmp(url, gb->NodeGetURL()) != 0 ) {
    gb->TrieInsertLinkURL(url); 
    if(gb->flags.show_links) {
      cout << " * " << url << endl;
    }
  } 
} 











