/** @file pvminterface.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include <string.h>
#include <cassert>
#include "pvminterface.h"

/// Starts the PVM interface and syncs all the tasks.
/**
 * Every task collects some information about itself which must
 * be sent to all others. The PVMInterface constructor's job is
 * to synchronize all tasks, broadcast this task's details and 
 * read all other tasks' details into the tasklist array.
 */
PVMInterface :: PVMInterface(const char *name, 
			     bool ismaster, int ntasks, 
			     int startid, int stopid) throw (runtime_error) {
#ifndef DISTRIBUTED_COMPUTING

  num_tasks = 0;
  is_master = true;

  last_message_tid = 0;
  last_message_size = 0;
  last_message_msgtag = 0;


#else
  assert(startid < stopid);

  int info, inum, bufid;

  is_master = ismaster;
  num_tasks = ntasks;
  tasklist = new PVMTaskInfo[num_tasks - 1];

  last_message_tid = 0;
  last_message_size = 0;
  last_message_msgtag = 0;

  // get our tid and package the info
  strncpy(myinfo.name, name, PVMINTERFACE_NAMELEN);
  myinfo.name[PVMINTERFACE_NAMELEN] = 0;
  myinfo.start_id = startid;
  myinfo.stop_id = stopid;
  myinfo.tid = pvm_mytid();
  if( myinfo.tid < 0 ) goto bailout;
  
  // join the synchronization group
  inum = pvm_joingroup("insync");
  if( inum < 0 ) goto bailout;

  // wait for all to join

  cerr << "pvm: waiting to sync with " << (num_tasks-1) << " other(s)" << endl;
  info = pvm_barrier("insync", num_tasks);
  if( info < 0 ) goto bailout;

  // now that everyone's here...
  // broadcast my details to everyone

  bufid = pvm_initsend(PvmDataDefault);
  if( bufid < 0 ) goto bailout;

  info = pvm_pkbyte((char*)&myinfo, sizeof(PVMTaskInfo), 1);
  if( info < 0 ) goto bailout;

  info = pvm_bcast("insync", PVMINTERFACE_SENDINFO);
  if( info < 0 ) goto bailout;

  // and read everyone else's details
  for(int k = 0; k < num_tasks - 1; k++) {
    bufid = pvm_recv( -1, PVMINTERFACE_SENDINFO);
    if( bufid < 0 ) goto bailout;
    info = pvm_upkbyte((char*)&tasklist[k], sizeof(PVMTaskInfo), 1);
    if( info < 0 ) goto bailout;
  } 

  // finally indicate standby mode
  JoinStandbyGroup();
  // done
  return;

 bailout:
  pvm_perror("PVMInterface constructor");
  pvm_exit();

  throw runtime_error("");

#endif
}

/// Closes down the PVM connection and notifies pvmd.
PVMInterface :: ~PVMInterface() {
#ifdef DISTRIBUTED_COMPUTING

  delete[] tasklist;
  pvm_exit();

#endif
}

bool PVMInterface :: MessagePending() {
#ifdef DISTRIBUTED_COMPUTING
  return (pvm_probe(-1,PVMINTERFACE_SENDCOMMAND) > 0);
#else
  return false;
#endif
}
/// Returns tasklist index of first task which contains the id number.
/// Returns -1 if not found.
int PVMInterface :: FindTask(uint32 id) {
#ifdef DISTRIBUTED_COMPUTING

  for(int i = 0; i < num_tasks - 1; i++) {
    if( (tasklist[i].start_id <= id) && (tasklist[i].stop_id >= id) ) {
      return i;
    }
  }
  return -1;

#else 
  return -1;
#endif
}

const char* PVMInterface :: Name(int k) {
#ifdef DISTRIBUTED_COMPUTING
  assert(k < num_tasks);
  return tasklist[k].name;
#else 
  return "";
#endif
}

bool PVMInterface :: AllStandby() throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING
  int size = pvm_gsize("standby");
  if( size < 0 ) {
    pvm_perror("PVMInterface::BroadcastCommand");
    pvm_exit();
    throw runtime_error("");
  } else {
    return (size >= num_tasks);
  }
#else
  return true;
#endif
}

void PVMInterface :: LeaveStandbyGroup() {
#ifdef DISTRIBUTED_COMPUTING
  cerr << "pvm: now leaving STANDBY" << endl;
  pvm_lvgroup("standby");
#endif
}

void PVMInterface :: JoinStandbyGroup() {
#ifdef DISTRIBUTED_COMPUTING
  cerr << "pvm: now joining STANDBY" << endl;
  pvm_joingroup("standby");
#endif
}

void PVMInterface :: SendCommand(int taskno, const char* commandline) throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING

  int info, bufid;

  assert( taskno >= 0);
  assert( taskno < num_tasks - 1);
  // don't send empty messages
  if( !*commandline ) {
    return;
  }

  bufid = pvm_initsend(PvmDataDefault);
  if( bufid < 0 ) goto bailout;

  info = pvm_pkstr(const_cast<char*>(commandline));
  if( info < 0 ) goto bailout;

  info = pvm_send(tasklist[taskno].tid, PVMINTERFACE_SENDCOMMAND);
  if( info < 0 ) goto bailout;

  return;

 bailout:
  pvm_perror("PVMInterface::SendCommand");
  pvm_exit();

  throw runtime_error("");

#endif
}
/// Sends a string command to every member of the STANDBY group
void PVMInterface :: BroadcastCommand(const char* commandline) throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING

  int info, bufid;

  // don't send empty messages
  if( !*commandline ) {
    return;
  }

  bufid = pvm_initsend(PvmDataDefault);
  if( bufid < 0 ) goto bailout;

  info = pvm_pkstr(const_cast<char*>(commandline));
  if( info < 0 ) goto bailout;

  info = pvm_bcast("standby", PVMINTERFACE_SENDCOMMAND);
  if( info < 0 ) goto bailout;

  return;

 bailout:
  pvm_perror("PVMInterface::BroadcastCommand");
  pvm_exit();

  throw runtime_error("");

#endif
}

/// returns the task number of the last received message
/// or -1 if not found
int PVMInterface :: LastMessageOriginator() {
#ifdef DISTRIBUTED_COMPUTING

  for(int k = 0; k < num_tasks - 1; k++) {
    if( last_message_tid == tasklist[k].tid ) {
      return k;
    }
  }
  return -1;

#else

  return 0;

#endif
}

/// Waits for and reads a command line sent by the master
/// If no command has arrived after a delay, returns with ""
void PVMInterface :: GetCommandWithTimeout(char* commandline) throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING
  struct timeval tmout;
  tmout.tv_sec = 1;
  tmout.tv_usec = 0;

  int bufid;
  if( (bufid = pvm_trecv(-1,PVMINTERFACE_SENDCOMMAND,&tmout)) > 0 ) {
    int info;
    info = pvm_bufinfo(bufid, &last_message_size, &last_message_msgtag, &last_message_tid);
    if( info < 0 ) goto bailout;

    info = pvm_upkstr(commandline);
    if( info < 0 ) goto bailout;
  }

  return;

 bailout:
  pvm_perror("PVMInterface::GetCommandWithTimeout");
  pvm_exit();

  throw runtime_error("");

#else
  // do nothing with command
#endif
}
/// Fills the countsarray with a list of occupied LeafNodes from another task
/**
 * The size parameter should be set to the maximum allowable size of the countsarray.
 * The returned value of size is the actual size of the countsarray.
 */
void PVMInterface :: GetLeafCounts(unsigned int* countsarray, uint32* size) throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING
  
  int bufid;
  uint32 maxsize = *size;
  if( (bufid = pvm_recv(-1,PVMINTERFACE_SENDLEAFUPD)) > 0 ) {
    int info;
    info = pvm_upkuint(size, 1, 1);
    if( info < 0 ) goto bailout;
    assert(maxsize >= *size);
    info = pvm_upkuint(countsarray, *size, 1);
    if( info < 0 ) goto bailout;
  }

  return;

 bailout:
  pvm_perror("PVMInterface::GetCommandWithTimeout");
  pvm_exit();

  throw runtime_error("");

#else
  // do nothing with command
#endif
}

void PVMInterface :: SendLeafCounts(int taskno, unsigned int* countsarray, uint32 size) throw (runtime_error) {
#ifdef DISTRIBUTED_COMPUTING

  int bufid, info;
  assert( taskno >= 0);
  assert( taskno < num_tasks - 1);

  bufid = pvm_initsend(PvmDataDefault);
  if( bufid < 0 ) goto bailout;

  info = pvm_pkuint(&size, 1, 1);
  info = pvm_pkuint(countsarray, size, 1);
  if( info < 0 ) goto bailout;
  
  info = pvm_send(tasklist[taskno].tid, PVMINTERFACE_SENDLEAFUPD);
  if( info < 0 ) goto bailout;

  return;

 bailout:
  pvm_perror("PVMInterface::SendLeafCounts");
  pvm_exit();

  throw runtime_error("");

#endif

}
