/** @file sampler.cc */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "sampler.h"
//#include "basictypes.h"
#include <math.h>
//#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

/// among other things, initializes the random number generator
WebSampler :: WebSampler(WebLinkGraph* g) throw (exception) {
  assert(g);

  graph = g;
  number_of_nodes = graph->size();
  allocation_list = new int32[number_of_nodes];
  memset(allocation_list, 0, number_of_nodes * sizeof(int32));

  ClearCounts();
  ClearScratch();

  r = gsl_rng_alloc(gsl_rng_taus);
  if( !r ) {
    cerr << "error: couldn't initialize the random number generator" << endl;
    throw exception();
  }

  eps = 1.0;
  xleaf = NULL;
}

/// clears occupation counts for all WebNode objects in the graph
void WebSampler :: ClearCounts() {
   for(WebNodeList::iterator i = graph->begin();
      i != graph->end(); i++ ) {
     (*i)->ClearOccupationCount();
   }
   last_run_size = 0;
   last_tagged_run_size = 0;
}

/// clears scratch area for all WebNode objects in the graph
void WebSampler :: ClearScratch() {
   for(WebNodeList::iterator i = graph->begin();
       i != graph->end(); i++ ) {
     (*i)->SetScratch(0);
   }
}

/// iterates through the WebNodeList and prints
/// the occupation counts for each WebNode. 
/**
 * Also
 * calculates a 95% confidence interval for each count. 
 * That's why it's a WebSampler member function, and not
 * a WebLinkGraph member function.
 */
void WebSampler :: PrintCounts(ostream& o) {
  char buf[1024];
  o << "# Sampler used: " << Name(buf) << endl;
  o << "# Number of samples produced: " << last_run_size << endl;
  o << "# node | occupation_count | 95% conf. interval" << endl;
  for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
    double k = (double)(*i)->OccupationCount();
    double eps = 1.96 * sqrt(k * (last_run_size - k)/last_run_size);
    o << (*i)->ID() << " " 
      << (*i)->OccupationCount() << " "
      << max(static_cast<int>(floor(k - eps)),0) << " " 
      << static_cast<unsigned int>(ceil(k + eps)) << endl;
  }
}

/// same as WebSampler::PrintCounts(), but only prints 
/// tagged documents.
void WebSampler :: PrintTagCounts(ostream& o) {
  char buf[1024];
  o << "# Sampler used: " << Name(buf) << endl;
  o << "# Number of (tagged) samples produced: " << last_tagged_run_size << endl;
  o << "# node | occupation_count | 95% conf. interval" << endl;
  for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++) {
    if( (*i)->Tagged(0) ) {
      double k = (double)(*i)->OccupationCount();
      double eps = 1.96 * sqrt(k * (last_tagged_run_size - k)/last_tagged_run_size);
      o << (*i)->ID() << " " 
	<< (*i)->OccupationCount() << " "
	<< max(static_cast<int>(floor(k - eps)),0) << " " 
	<< static_cast<unsigned int>(ceil(k + eps)) << endl;
    }
  }
}

/// Clears the list of starting places.
void WebSampler :: ClearAllocForward() {
  assert(number_of_nodes > 0);
  memset(allocation_list, 0, number_of_nodes * sizeof(int32));
}

/// Adds num to the starting allocation for id
void WebSampler :: IncrementAllocForward(uint32 k, int32 num) {
  assert( static_cast<unsigned int>(k) < number_of_nodes);
  allocation_list[k] += num;
}

void WebSampler :: SimulateAllocForward() {
  // now for each starting place, simulate k steps
  // forward, where k ~ Geom(epsilon)
  uint32 j = 0;
  for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++, j++ ) {
    for(int k = allocation_list[j]; k > 0; k--) {

      WebNodePtr x = (*i);
      assert(x);
      while( (gsl_rng_uniform(r) > eps) &&
	     x && (x->NumberOfValidToLinks() > 0)) {
	x = QEvolveFrom(x);
      }

      if( x ) {
	x->IncrementOccupationCount();
      } else if( xleaf ) {
	xleaf->IncrementOccupationCount();
      }

    }
  }
}

/// samples n random variables from the page ranking distribution
/**
 * Each of the sampled variables is
 * generated by a uniform on the web link graph, followed
 * by a geometric number (mean eps) of transitions using
 * the virtual function QEvolveFrom(). The final WebNodePtr value is 
 * recorded in the WebNode's occupation_number data member.
 */
void WebSampler :: SimulateForward(long n) {
  assert(graph);
  assert(allocation_list);
  assert( n > 0 );

  assert(number_of_nodes > 0);

  ClearAllocForward();

  // generate n starting places
  for(long i = 0; i < n; i++)
    allocation_list[(int32)gsl_rng_uniform_int(r, number_of_nodes)]++;

  SimulateAllocForward();

  last_run_size += n;
}

/// samples from part of the page ranking distribution by rejection 
/// sampling. 
/** 
 * A total of n candidates are generated, not all
 * of which are eventually accepted. 
 * Only candidates which are accepted show up in the relevant
 * WebNode's occupation_count. 
 * Before calling this, 
 * make sure to call Grapher::BuildFromSets(), whithout which
 * this algorithm will bomb. 
 */
void WebSampler :: TaggedSimulateForward(long n, const uint32 *fromsetsize) {
  assert(graph);
  assert(allocation_list);
  assert( n > 0 );

  assert(number_of_nodes > 0);

  // figure out how many must start in a FromSet, and 
  // how many start anywhere in the graph
  long num_less_than_K = 0;
  long num_greater_than_K = 0;
  double threshold = pow(1 - eps, TAG_NUMBER_OF_BITS);
  for(long i = 0; i < n; i++) {
    if( gsl_rng_uniform(r) > threshold ) {
      num_less_than_K++;
    } else {
      num_greater_than_K++;
    }
  }
  // deal with chains whose number of transitions exceeds K
  // this code is nealy identical to SimulateForward
  {
    cerr << "info: tagged simulation uses " << num_greater_than_K << " global candidates (" 
	 << (100 * num_greater_than_K)/n << "%)"<< endl;
    long num_accepted = 0;
    // generate n starting places
    memset(allocation_list, 0, number_of_nodes * sizeof(int32));
    for(long i = 0; i < num_greater_than_K; i++)
      allocation_list[(int32)gsl_rng_uniform_int(r, number_of_nodes)]++;

    // now for each starting place, simulate K+s steps
    // forward, where s ~ Geom(epsilon)
    uint32 j = 0;
    for(WebNodeList::iterator i = graph->begin();
	i != graph->end(); i++, j++ ) {
      for(int k = allocation_list[j]; k > 0; k--) {
	
	WebNodePtr x = (*i);
	assert(x);
	// step through K mandatory transitions
	for(int t = 0; (t < TAG_NUMBER_OF_BITS) && (x->NumberOfValidToLinks() > 0); t++) {
	  x = QEvolveFrom(x);
	}
	// now generate the others
	while( (gsl_rng_uniform(r) > eps) && (x->NumberOfValidToLinks() > 0)) {
	  x = QEvolveFrom(x);
	}
	// accept only if the result is tagged
	if( x->Tagged(0) ) {
	  x->IncrementOccupationCount();
	  last_tagged_run_size++;
	  num_accepted++;
	}
      }
    }
    cerr << "info: tagged simulation has produced " << num_accepted << " global samples" << endl;
  }

  // deal with chains whose number of transitions is less than K
  {
    cerr << "info: tagged simulation uses " << num_less_than_K << " local candidates" << endl;
    long num_accepted = 0;
    /// generate num_less_than_K starting points

    for(int k = 0; k < TAG_NUMBER_OF_BITS; k++) {
      probabilities[k] = static_cast<double>(fromsetsize[k]) * eps * pow(1 - eps, k);
      allocated[k] = 0;
      fromsetsize_remaining[k] = fromsetsize[k]; // will be decremented below
    }
    gsl_ran_discrete_t *grd = gsl_ran_discrete_preproc(TAG_NUMBER_OF_BITS, probabilities);
    for(long l = 0; l < num_less_than_K; l++ ) {
      // allocated[k] contains the number of starting poinst
      // allocated to the kth FromSet
      allocated[gsl_ran_discrete(r,grd)]++;
    }
    gsl_ran_discrete_free(grd);

    // now step through the web graph
    for(WebNodeList::iterator i = graph->begin(); i != graph->end(); i++ ) {
      assert(*i);
      for(int k = 0; k < TAG_NUMBER_OF_BITS; k++) {
	if( (*i)->Tagged(k) ) {
	  // this node belongs to the kth FromSet
	  fromsetsize_remaining[k]--;
	  assert(fromsetsize_remaining[k] >= 0);

	  // see how many times it should be chosen as a starting point
	  double p = 1.0/static_cast<double>(fromsetsize_remaining[k]);
	  unsigned int num_starts = gsl_ran_binomial(r, p, allocated[k]);

	  // now that we have the number of starting chains, ie
	  // num_starts ~ Binomial(allocated[k], p),
	  // we iterate each chain exactly k times
	  for(uint32 c = 0; c < num_starts; c++) {
	    WebNodePtr x = (*i);
	    int s = k;
	    while( (s > 0) && (x->Tagged(s)) ) {
	      // if x fails to be tagged(s), then 
	      // there is no hope of reaching the tagged(0) set
	      x = QEvolveFrom(x);
	      s--;
	    }
	    // accept only if x belongs to tagged set
	    if( (s == 0) && (x->Tagged(0)) ) { // test s==0 is probably unnecessary
	      x->IncrementOccupationCount();
	      last_tagged_run_size++;
	      num_accepted++;
	    }
	  }

	  // now that we've run the chains, remove them
	  // from the total allocated number
	  allocated[k] -= num_starts; 
	  assert(allocated[k] >= 0);
	}
      }
    }
    cerr << "info: tagged simulation has produced " << num_accepted << " local samples" << endl;
  }
}

//-------------------------------------//

PageRankSampler :: PageRankSampler(WebLinkGraph* agraph): WebSampler(agraph) {
  eps = 0.5;
}

void PageRankSampler :: SetParameters(double epsilon) {
  assert(graph);
  assert( (epsilon > 0) && (epsilon < 1));
  eps = epsilon;
  ClearCounts();
}

/// Returns the effect of a single random transition.
/// If result is NULL, this means the new state is a LeafNode
inline WebNodePtr PageRankSampler :: QEvolveFrom(WebNodePtr x) {

  int nx = x->NumberOfValidToLinks() + x->NumberOfLeafLinks();
  if( nx == 0 ) {
    return x;
  }

  register int where = 
    gsl_rng_uniform_int(r, nx);
  if( where < x->NumberOfValidToLinks() ) {
    return x->ValidToLink(where);
  } else {
    xleaf = x->ValidLeafLinkDirectly(where); 
    return NULL;
  }

}

char *PageRankSampler :: Name(char *buf) {
  sprintf(buf, "PageRank, epsilon = %f", eps);
  return buf;
}

//-------------------------------------//

DateBiasedPageRankSampler :: DateBiasedPageRankSampler(WebLinkGraph* agraph): WebSampler(agraph) {
  eps = 0.5;
  lam = 0.01;
  daterange = kuint16max;
  lamhat = lam/daterange;

  // this is for safety, don't remove
  assert(sizeof(DBScratchStruct) == sizeof(ScratchStruct));
}

void DateBiasedPageRankSampler :: SetParameters(double epsilon, double lambda, uint16 dr) {
  assert(graph);
  assert( (epsilon > 0) && (epsilon < 1));
  eps = epsilon;
  lam = lambda;
  assert(dr > 0);
  daterange = dr;
  lamhat = lam/daterange;
  ClearCounts();
}


/// Returns the effect of a single random transition.
/// If result is NULL, this means the new state is a LeafNode
inline WebNodePtr DateBiasedPageRankSampler :: QEvolveFrom(WebNodePtr x) {

  int nx = x->NumberOfValidToLinks() + x->NumberOfLeafLinks();
  if( nx == 0 ) {
    return x;
  }

  register DBScratchStruct scratch(x->Scratch());
  // work out total mass if needed
  if( !scratch.ss ) {
    scratch.mass = 0.0;
    for(register int t = 0; t < x->NumberOfValidToLinks(); t++) {
      scratch.mass += exp(-lamhat * max(x->Date() - x->ValidToLink(t)->Date(), 0));
    }
    for(register int t = 0; t < x->NumberOfLeafLinks(); t++) {
      scratch.mass += exp(-lamhat * max(x->Date() - x->ValidLeafLink(t)->Date(), 0));
    }
    x->SetScratch(scratch.ss);
  }

  register float tmass = 0.0;

  for(register int t = 0; t < x->NumberOfValidToLinks(); t++) {
    register float xi = gsl_rng_uniform(r);
    register float p = exp(-lamhat * max(x->Date() - x->ValidToLink(t)->Date(),0) );
    if( p + tmass * xi >= scratch.mass * xi ) {
      return x->ValidToLink(t);
    } else {
      tmass += p;
    }
  }

  for(register int t = 0; t < x->NumberOfLeafLinks(); t++) {
    register float xi = gsl_rng_uniform(r);
    register float p = exp(-lamhat * max(x->Date() - x->ValidLeafLink(t)->Date(),0) );
    if( p + tmass * xi >= scratch.mass * xi ) {
      xleaf = x->ValidLeafLink(t);
      return NULL;
    } else {
       tmass += p;
    }
  }

  // if we're here, there was a precision problem
  // log it and choose a link at random like in PageRank
  cerr << "warning: precision difficulties in DateBiasedPageRankSampler::QEvolveFrom()" << endl;

  register int where = 
    gsl_rng_uniform_int(r, nx);
  if( where < x->NumberOfValidToLinks() ) {
    return x->ValidToLink(where);
  } else {
    xleaf = x->ValidLeafLinkDirectly(where); 
    return NULL;
  }

}


char *DateBiasedPageRankSampler :: Name(char *buf) {
  sprintf(buf, "DateBiasedPageRank, epsilon = %f, lambda = %f", eps, lam);
  return buf;
}

///
/// old implementation
///
// /// Returns the effect of a single random transition.
// /// If result is NULL, this means the new state is a LeafNode
// inline WebNodePtr DateBiasedPageRankSampler :: QEvolveFrom(WebNodePtr x) {

//   int nx = x->NumberOfValidToLinks() + x->NumberOfLeafLinks();
//   if( nx == 0 ) {
//     return x;
//   }

//   register DBScratchStruct scratch(x->Scratch());
//   // work out normalising constant if needed
//   if( !scratch.db.filled ) {
//     register uint16 maxdate = 0;
//     for(register int t = 0; t < x->NumberOfValidToLinks(); t++) {
//       maxdate = max(maxdate, x->ValidToLink(t)->Date());
//     }
//     for(register int t = 0; t < x->NumberOfLeafLinks(); t++) {
//       maxdate = max(maxdate, x->ValidLeafLink(t)->Date());
//     }
//     scratch.db.minvalue = x->Date() - maxdate;
//     scratch.db.filled = 1;
//     x->SetScratch(scratch.ss);
//   }

//   // rejection sampler
//   register int where = gsl_rng_uniform_int(r, nx);

//   while(1) {
//     register uint16 d = (where < x->NumberOfValidToLinks()) ? 
//       x->ValidToLink(where)->Date() : 
//       x->ValidLeafLinkDirectly(where)->Date();
//     if( lam * (x->Date() - d - scratch.db.minvalue) 
// 	<= gsl_ran_exponential(r,1.0) ) {
//       break;
//     } else {
//       where = gsl_rng_uniform_int(r, nx);
//     }
//   }
    
//   if( where < x->NumberOfValidToLinks() ) {  
//     return x->ValidToLink(where);
//   } else {
//     xleaf = x->ValidLeafLinkDirectly(where);
//     return NULL;
//   }
// }

//-------------------------------------//

TruncatedKleinbergSampler :: TruncatedKleinbergSampler(WebLinkGraph* agraph): WebSampler(agraph) {
  eps = 0.5;
}

void TruncatedKleinbergSampler :: SetParameters(double epsilon, ktype what) {
  assert(graph);
  assert( (epsilon > 0) && (epsilon < 1));
  eps = epsilon;
  which = what;
  ClearCounts();
}

/// Returns the effect of a single random transition.
/// Not designed to work with leaf nodes
inline WebNodePtr TruncatedKleinbergSampler :: QEvolveFrom(WebNodePtr x) {

  int nx;

  switch(which) {
  case hubs:
    nx = x->NumberOfValidToLinks();
    if( nx > 0 ) {
      // forward
      register int where = gsl_rng_uniform_int(r, nx);
      x = x->ValidToLink(where);
      // backward
      nx = x->NumberOfValidFromLinks();
      if( nx > 0 ) {
	where = gsl_rng_uniform_int(r, nx);
	return x->ValidFromLink(where);
      } else {
	return x;
      }
    } else {
      return x;
    }
    break;
  case auth:
    nx = x->NumberOfValidFromLinks();
    if( nx > 0 ) {
      // backward
      register int where = gsl_rng_uniform_int(r, nx);
      x = x->ValidFromLink(where);
      // forward
      nx = x->NumberOfValidToLinks();
      if( nx > 0 ) {
	where = gsl_rng_uniform_int(r, nx);
	return x->ValidToLink(where);
      } else {
	return x;
      }
    } else {
      return x;
    }
    break;
  default:
    return x;
  }

}

char *TruncatedKleinbergSampler :: Name(char *buf) {
  sprintf(buf, "Truncated Kleinberg, epsilon = %f %s", eps,
	  (which == hubs) ? "(hubs)" : "(auth)");
  return buf;
}
