/** @file simplehash.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _SIMPLE_HASH_H_
#define _SIMPLE_HASH_H_

#include "basictypes.h"
#include <stddef.h>
#include <string.h>
#include <cassert>
#include <stdexcept>

/// A structure used by SimpleHashTable.
template<class R>
struct SimpleHashPair {
  ptrdiff_t key;
  R value;
};

/// A simple "linear probing and insertion" hashtable.
/**
 * This simplistic hashtable is designed as a 
 * replacement for the STL hashtable, which uses too much memory.
 * The hashtable implements algorithm L in Knuth, Vol.3, p.526 and
 * also uses Knuth's hash function.
 */
template<class R>
class SimpleHashTable
{

 public:
  SimpleHashTable(uint32 size);
  uint32 IntHash(uint32 key);
  R Find(ptrdiff_t key);
  bool Insert(ptrdiff_t key, R value) throw (overflow_error);
  inline uint32 GetCount() { return count; };
  inline uint32 Size() { return size_; };
  void Clear();

 private:
  uint32 size_;
  uint32 count;
  SimpleHashPair<R> *table;

};

template<class R>
SimpleHashTable<R> :: SimpleHashTable(uint32 size) {
  size_ = size;
  count = 0;
  table = new SimpleHashPair<R>[size_];
  memset(table, 0, sizeof(SimpleHashPair<R>) * size_);
}

template<class R>
void SimpleHashTable<R> :: Clear() {
  count = 0;
  memset(table, 0, sizeof(SimpleHashPair<R>) * size_);
}

template<class R>
R SimpleHashTable<R> :: Find(ptrdiff_t key) {
  uint32 j = IntHash((uint32)key);
  uint32 jj = j % size_;
  while( table[jj].value != (R)0 ) {
    if(table[jj].key == key) {
      return table[jj].value;
    }
    jj = (jj > 0) ? jj - 1 : size_ - 1;
  }
  return (R)0;
}

/// Inserts a key,value pair into the hash. Returns false if we 
/// try to insert a duplicate key, and returns an exception 
/// if hash is full
template<class R>
bool SimpleHashTable<R> :: Insert(ptrdiff_t key, R value) throw (overflow_error) {

  if( count < size_ ) {
    uint32 j = IntHash((uint32)key);
    uint32 jj = j % size_;
    while( table[jj].value != (R)0 ) {
      if(table[jj].key == key) 
	{ return false; }
      jj = (jj > 0) ? jj - 1 : size_ - 1;
    }
    table[jj].key = key;
    table[jj].value = value;
    count++;
    return true;
  } else {
    cerr << "SimpleHashTable overflow" << endl;
    throw overflow_error("");
  }
}

template<class R>
uint32 SimpleHashTable<R> :: IntHash(uint32 key) {
// due to thomas wang, a simplification of bob jenkins' 96 DDJ hash
//   key += ~(key << 15);
//   key ^=  (key >>> 10);
//   key +=  (key << 3);
//   key ^=  (key >>> 6);
//   key += ~(key << 11);
//   key ^=  (key >>> 16);

// knuth's golden ratio of 2^32
  key = (key >> 3) * 2654435761UL;
  return key;
}

#endif
