/** @file talker.cc 
 * This file contains the member functions for the
 * class Talker.
 */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */
/**
 * @mainpage 
 *
 * @section int Introduction
 *
 * The MarkovPR project consists of a collection of programs and datasets
 * designed to permit the
 * study of web document page ranking schemes. These programs consist of 
 *   - A dataset, provided by Google, Inc, consisting of 
 *     compressed, preprocessed repository files pprepos.*.bz2,
 *     to be placed in the subdirectory 
 *     @htmlonly <a href="../../data">data/</a> @endhtmlonly
 *   - A ripper program, based upon the skeleton provided by Google, Inc, which builds
 *     a link graph G of all the documents in the repository, and runs various Markov chain
 *     algorithms on G. The %main entry point for ripper is in the file ripper.cc
 *   - A client program jack, which permits users to interact with ripper. The %main entry
 *     point for jack is the file jack.cc
 *
 * A overview of the full project is presented 
 * @htmlonly <a href="../googlerip.ps.gz">here</a>. @endhtmlonly
 *
 * @section cl Copyrights and Licenses
 *
 * The distributed source code and programs are partly based upon the code provided freely by 
 * Google, Inc., on the occasion of their First Annual Programming Contest, for research
 * and noncommercial purposes. See their 
 * @htmlonly <a href="../../LICENSE.Google">LICENSE.Google</a>.@endhtmlonly
 * While Google retains the copyright for its supplied source files, including all modifications
 * to them, the remaining source files (clearly marked) are distributed under the 
 * @htmlonly <a href="../../GPL">GPL</a>.@endhtmlonly
 *
 * The dataset required by the ripper program must be obtained separately from Google, Inc.
 * At the time of writing, the necessary files can be downloaded directly. See instructions
 * in the file @htmlonly <a href="../../README.Google">README.Google</a>.@endhtmlonly
 *
 * @section rip Overview of Ripper
 *
 * @subsection ripsub1 General
 *
 * The ripper reads the (uncompressed) preparsed repository files on STDIN and builds
 * a link graph in-memory. Once the graph is built, the ripper enters an interactive
 * mode which allows it to communicate with the jack program via two pipes.
 * 
 * The program can be compiled by issuing the command @verbatim make all@endverbatim from within 
 * the src directory. An optimized version without debug information is build by typing
 * @verbatim make optimized @endverbatim The full documentation can be regenerated
 * by typing @verbatim make doc @endverbatim which requires the programs 
 * @htmlonly <a href="http://www.doxygen.org">doxygen</a>@endhtmlonly and 
 * @htmlonly <a href="http://www.research.att.com/sw/tools/graphviz/">graphviz</a>.@endhtmlonly
 *
 * After compilation, a typical invocation would be 
 * @verbatim bzcat data/pprepos.*.bz2 | src/ripper - @endverbatim
 * Various options can be included on the ripper command line. These are listed by running
 * the command @verbatim ripper --help @endverbatim
 *
 * The ripper displays some status information on STDERR. For example, once all repositories
 * have been read, it will say so. The user can then invoke the program jack to enter ripper's
 * interactive mode.
 *
 * @subsection ripsub2 Web link graph construction
 *
 * The structure of the ripper is roughly the following: Ripper is responsible for reading
 * repositories and invoking callback functions for every encountered web document. The relevant
 * callbacks for building the web link graph are handled by the class GraphParseHandler. The 
 * conversion of this document into a WebNode is the purpose of the class GraphBuilder. Among
 * its data members are the Trie, which allows all encountered URLs to be stored and searched,
 * and the nodetable, which is a SimpleHashTable<WebNodePtr> and allows a specific 
 * WebNode to be found
 * from its URL (this is necessary since a WebNode uses an ID number for
 * identification, *not* a URL directly
 * for space reasons).
 * Once GraphBuilder has produced the WebLinkGraph, the latter is passed onto Talker, which 
 * instantiates several WebSampler classes. Talker's main purpose is to read pipes for commands to
 * execute, this being mainly running WebSampler classes and supporting functions.
 *
 * @subsection ripsub3 Parallel computation
 *
 * The ripper/jack programs can be used in conjunction with PVM to perform simulations
 * in parallel. This requires the 
 * @htmlonly <a href="www.csm.ornl.gov/pvm/pvm_home.html">PVM</a> @endhtmlonly
 * libraries and include files to be installed. You
 * can compile the ripper/jack programs with the command
 * @verbatim make distributed @endverbatim
 * which requires also that the PVM_ROOT environment variable be set (see installation 
 * instructions for PVM). Once the programs are compiled, you can run several copies
 * of ripper on different machines. Only one master is allowed to communicate with jack. 
 * Here is an example .pvmrc file for two machines ender and enigma:
 * @verbatim
add enigma
conf
ps -axl
spawn -enigma ripper --pvm_slave 2 --name darth --temp_dir tmp/ data/pprepos.??
spawn -ender ripper --pvm_master 2 --name palpatine proj/cpp/google_contest/data/pprepos.??
@endverbatim
 * This assumes that on the machine ender, the dataset files (uncompressed) can be found
 * in $HOME/proj/cpp/google_contest/data. This directory is mounted remotely on enigma 
 * as the directory $HOME/data. Also, temporary files will be created in /tmp on ender,
 * which should be remotely mounted (r/w) on enigma as $HOME/tmp.
 *
 * To control the master, you should execute on ender the command @verbatim
jack --name palpatine
@endverbatim
 * (but wait until palpatine has finished processing the dataset).
 *
 * @subsection ripsub4 Miscellaneous ripper options
 *
 * The ripper can also be used to simply extract information from the repository files.
 * For example, typing @verbatim
bzcat ../data/pprepos.*.bz2 | ripper --catlinks -
@endverbatim
 * extracts the link urls to STDOUT. Other options which perform similarly include 
 * --cat, --caturl, --catdate. 
 *
 * @section jack Overview of Jack
 *
 * The program jack communicates with ripper via two pipes, whose default names are 
 * /tmp/ripper.input and /tmp/ripper.output respectively. Once invoked, the user can
 * type simple interactive commands which ripper will execute. These commands include
 * running various Markov chains on the link graph, saving page ranking calculations
 * to a file, etc. The jack program accepts two command line options which
 * behave identically to their ripper namesakes, namely --temp_dir and --name. You
 * should use these options if you invoke ripper with them also.
 */
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include "talker.h"
#include <gsl/gsl_randist.h>

extern int num_docs_processed;
static char helptext[] = "Available commands are:\n"
"quit - exits the program\n"
"help - displays this text\n"
"info mem - displays memory usage statistics\n"
"\n"
"info sampler - displays information about the selected sampler\n"
"set sampler pagerank epsilon - selects ordinary PageRank algorithm\n"
"set sampler dateb epsilon lambda - selects date biased PageRank algorithm\n"
"set sampler tkleinberg epsilon hubs|auth - selects truncated Kleinberg algorithm\n"
"run sampler n - runs the current Markov chain sampler for n iterations\n"
"load tags filename - tags all documents listed in filename\n"
"run tagsampler n - runs the current sampler on tagged documents only\n"
"reset counts - resets the Markov chain occupation counts\n"
"save counts filename - prints the occupation counts in filename\n"
"save tagcounts filename - prints the occupation counts for tagged documents only\n"
"\n"
"info graph - displays some statistics about the web graph\n"
"save graph filename - prints the web link graph in filename\n"
"";

Talker::Talker(const char* ripper_name, const char* ripper_tempdir, 
	       WebLinkGraph *ripper_web, GraphBuilder *ripper_gb,
	       bool pvm_is_master, int pvm_numtasks) throw (exception) {
  name = ripper_name;
  tempdir = ripper_tempdir;
  web = ripper_web;
  gb = ripper_gb;

  r = gsl_rng_alloc(gsl_rng_taus);
  if( !r ) {
    cerr << "error: couldn't initialize the random number generator" << endl;
    throw exception();
  }

  // set up some default samplers
  prsampler = new PageRankSampler(web);
  prsampler->SetParameters(0.1);

  dbsampler = new DateBiasedPageRankSampler(web);
  dbsampler->SetParameters(0.1,0.1,gb->HighestDate() - gb->LowestDate());

  tksampler = new TruncatedKleinbergSampler(web);
  tksampler->SetParameters(0.1,hubs);

  defaultsampler = NULL;

  leaflist = new LeafNodePtrList();

  pvminterface = NULL;

  assert(gb);
  pvminterface = new PVMInterface(name, 
				  pvm_is_master, pvm_numtasks, 
				  gb->LowestID(), gb->HighestID());

  if( pvm_numtasks > 0 ) {
    
    if( pvm_is_master ) {
      SetupFIFOs();
    }
    
    nodequiv << tempdir << "ripper.";
    if( *name ) {
      nodequiv << name << ".";
    }
    nodequiv << "nodes.equiv";
    remove(nodequiv.str().c_str());

    // build the leafnodes
    cerr << "info: now building leafnodes" << endl; 
    for(int k = 0; k < pvminterface->NumberOfOtherTasks(); k++) {
      ostringstream fname;
      fname << tempdir << "ripper." << pvminterface->Name(k) << ".index";
      ifstream f(fname.str().c_str());
      if( !f.fail() ) {
	if( !LoadLeaves(f) ) {
	  cerr << "warning: index file was not properly loaded" << endl;
	}
	f.close();
      } else {
	cerr << "error: failed to open index file " << fname << endl;
      }
    }
  } else {
    SetupFIFOs();
  }
}

Talker::~Talker() {
  cerr << "info: quitting." << endl;
  RemoveFIFOs(); 
  delete pvminterface;
  delete prsampler;
  delete dbsampler;
  delete leaflist;
}

/// Creates input/output FIFOs if they don't exist
/// Also deletes nodes.equiv if it exists
void Talker::SetupFIFOs() throw (exception) {
  cerr << "info: setting up FIFOs" << endl;

  inputfifo << tempdir << "ripper.";
  if( *name ) {
    inputfifo << name << ".";
  }
  inputfifo << "input";

  if( mkfifo(inputfifo.str().c_str(), S_IWUSR|S_IRUSR ) < 0 ) {
    cerr << "couln't create FIFO, or FIFO already exists [" << inputfifo.str() << "]" << endl;
    throw exception();
  }

  outputfifo << tempdir << "ripper.";
  if( *name ) {
    outputfifo << name << ".";
  }
  outputfifo << "output";

  if( mkfifo(outputfifo.str().c_str(), S_IWUSR|S_IRUSR ) < 0 ) {
    cerr << "couln't create FIFO, or FIFO already exists [" << outputfifo.str() << "]" << endl;
    throw exception();
  }

}

/// Unlinks the FIFO files
void Talker::RemoveFIFOs() {
  remove(inputfifo.str().c_str());
  remove(outputfifo.str().c_str());
}

/// Reads and executes commands through a couple of pipes
/**
 * This function executes interactive commands.
 * Input is read from /tmp/ripper.input
 * and resposes are written to /tmp/ripper.output
 */
void Talker::Talk() throw (exception) {

  ifstream in;
  ofstream out;

  defaultsampler = prsampler;

  if( (pvminterface->NumberOfTasks() == 0) || pvminterface->IsMaster() ) {
    cerr << "info: now listening for jack..." << endl;

    in.open(inputfifo.str().c_str());
    out.open(outputfifo.str().c_str(),ios::nocreate|ios::app);
    if( !in || !out ) {
      cerr << "error: cannot open FIFO " << endl;
      throw exception();
    }

    // tell ripper-client that all's well
    // communication with ripper-client uses
    // the \r character to denote the end of a message
    // always call flush after a message
    out << '\r'; 
    out.flush(); 

    cerr << "info: now entered interactive mode." << endl;
  }


  char commandline[1024];
  bool quit = false;

  while(!quit) {
    assert(pvminterface->AllStandby());
    memset(commandline,0,1024);

    // master reads input from jack and broadcasts command line
    if( (pvminterface->NumberOfTasks() == 0) || pvminterface->IsMaster() ) {
      // read
      in.getline(commandline, 1024);
      cerr << "info: received command [" << commandline << "]" << endl;
      // simplify
      if( *commandline == 0 ) {
	out << '\r';
	out.flush();
	continue;
      }
    }

    // get the command line from broadcast and process it.
    // continue with this until we receive a finish message
    // empty commands are not processed, but master will
    // check if everyone has finished their computations
    do {

      if( *commandline ) {
	// do work
	if(pvminterface->NumberOfTasks() > 0) {
	  cerr << "pvm: received message [" << commandline << "]" <<  endl;
	}
	pvminterface->LeaveStandbyGroup();
	quit = ProcessCommand(commandline, out);
	pvminterface->JoinStandbyGroup();
      } else if( (pvminterface->NumberOfTasks() > 0) && 
		 pvminterface->IsMaster() &&
		 pvminterface->AllStandby() ) {
	// master tells slaves to stop computing
	cerr << "pvm: master says finish_now" << endl;
	pvminterface->BroadcastCommand("pvmcontrol_finish_now");
	break; // master doesn't receive broadcast
      }

      // if no command received, this leaves commandline "as is"
      pvminterface->GetCommandWithTimeout(commandline);

    } while( !quit && 
	     (pvminterface->NumberOfTasks() > 0)  && 
	     (strcmp(commandline,"pvmcontrol_finish_now") != 0) );

    // master notifies jack that command is done
    if( (pvminterface->NumberOfTasks() == 0) || pvminterface->IsMaster() ) {
      out << '\r';
      out.flush();
    } else { // must be slave
      cerr << "pvm: finished" << endl;
    }
  }

  // close FIFOs
  if( (pvminterface->NumberOfTasks() == 0) || pvminterface->IsMaster() ) {
    in.close();
    out.close();
  }
}

/// Used for sorting in PropagateLeafCounts()
bool greater_LeafCountsStruct(const LeafCountsStruct x, const LeafCountsStruct y) {
  return (x.id > y.id);
}


/// Sends the LeafNode's occupation counts to other processes
/// and clears the counts. 
/**
 * This procedure is interruptible by pending messages. 
 *
 * Exception caveat:
 * Because we want to cast the structure LeafCountsStruct as
 * a consecutive pair of unsigned integers, we throw an invalid_argument
 * exception in case sizeof(unsigned int) != sizeof(uint32) for this 
 * compiler. Not very portable, but less work :-)
 */
void Talker::PropagateLeafCounts() throw (invalid_argument) {

  LeafCountsStruct* counts = new LeafCountsStruct[MAXLEAFMESSAGE_SIZE];

  int taskno;
  long num_particles;
  do {
    taskno = -1;
    num_particles = 0;
    uint32 c = 0;
    for(LeafNodePtrList::iterator i = leaflist->begin(); 
	i != leaflist->end(); i++) {
      if( (*i)->OccupationCount() > 0 ) {
	if( taskno < 0 ) {
	  if( (taskno = pvminterface->FindTask((*i)->ID())) > -1 ) {
	    counts[c].id = (*i)->ID();
	    counts[c].occupation_count = (*i)->OccupationCount();
	    num_particles += (*i)->OccupationCount();
	    (*i)->ClearOccupationCount();
	    c++;
	  }
	} else if ( ((*i)->ID() >= pvminterface->StartID(taskno)) && 
		    ((*i)->ID() <= pvminterface->StopID(taskno)) ) {
	    counts[c].id = (*i)->ID();
	    counts[c].occupation_count = (*i)->OccupationCount();
	    num_particles += (*i)->OccupationCount();
	    (*i)->ClearOccupationCount();
	    if( ++c == (MAXLEAFMESSAGE_SIZE - 1)) {
	      break;
	    }
	}
      }
    }
    
    if( taskno > -1 ) {
      // sort counts in decreasing id order.
      cerr << "pvm: sending " << c << " LeafNodes to " << pvminterface->Name(taskno) 
	   << " (" << num_particles << " particles)" << endl;
      sort(counts,counts+c,greater_LeafCountsStruct);
      // send message
      if( sizeof(unsigned int) == sizeof(uint32) ) {
	uint32 size = sizeof(LeafCountsStruct)*c/sizeof(unsigned int);
	assert(size > 0);
	pvminterface->SendCommand(taskno, "pvm run sampler"); // tell task we're sending leaf counts
	pvminterface->SendLeafCounts(taskno, (unsigned int*)counts, size);
      } else {
	throw invalid_argument("sizeof(unsigned int) != sizeof(uint32)");
      }
    }

  } while( (taskno > -1) && !pvminterface->MessagePending());

  delete[] counts;
}

bool Talker::ProcessCommand(char *commandline, ostream& out) {
  bool quit = false;
  char *t;

  switch( GetToken(commandline, out, "try help", 9,
		   "quit", "help", "info", "save", 
		   "reset", "set", "run", "load", "pvm") ) {
  case 0: // quit
    if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
      pvminterface->BroadcastCommand("quit");
    }
    quit = true;
    break;
  case 1: // help
    out << helptext;
    break;
  case 2: // info
    switch( GetToken(NULL, out, "try info [mem|sampler|graph]", 3,
		     "mem", "sampler", "graph") ) {
    case 0: // info mem
      PrintStatistics(out);
      break;
    case 1: // info sampler
      char buf[100];
      out << "Current sampler is: " << defaultsampler->Name(buf) << endl;
      out << "Total perfect samples produced so far: " << static_cast<int>(defaultsampler->LastRunSize()) << endl;
      out << "Total tagged samples produced so far: " << static_cast<int>(defaultsampler->LastTaggedRunSize()) << endl;
      break;
    case 2: // info graph
      PrintStatisticsGraph(out);
      web->StatisticsFromSets(out);
      break;
    }
    break;
  case 3: // save
    switch( GetToken(NULL, out, "try save [graph|counts]", 3,
		     "graph", "counts", "tagcounts") ) {
    case 0: // save graph
      if( (t = GetTokenVar(NULL, out, 
			   "try save graph filename")) ) {
	ofstream g(t);
	if( !g.fail() ) {
	  if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	    ostringstream s;
	    s << "pvm save graph " << t;
	    pvminterface->BroadcastCommand(s.str().c_str());
	  }
	  web->PrintWebGraph(g);
	  out << "graph stored in " << t <<  endl;
	  g.close();
	} else {
	  out << "Error: could not open file [" << t << "]" << endl;
	}
      }
      break;
    case 1: // save counts
      if( (t = GetTokenVar(NULL, out, 
			   "try save counts filename")) ) {
	ofstream g(t);
	if( !g.fail() ) {
	  if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	    ostringstream s;
	    s << "pvm save counts " << t;
	    pvminterface->BroadcastCommand(s.str().c_str());
	  }
	  defaultsampler->PrintCounts(g);
	  out << "counts stored in " << t <<  endl;
	  g.close();
	} else {
	  out << "Error: could not open file [" << t << "]" << endl;
	}
      }
      break;
    case 2: // save tagcounts
      if( (t = GetTokenVar(NULL, out, 
			   "try save tagcounts filename")) ) {
	ofstream g(t);
	if( !g.fail() ) {
	  defaultsampler->PrintTagCounts(g);
	  out << "tagcounts stored in " << t <<  endl;
	  g.close();
	} else {
	  out << "Error: could not open file [" << t << "]" << endl;
	}
      }
      break;
    }
    break;
  case 4: // reset
    switch( GetToken(NULL, out, "try reset counts", 2,
		     "counts", "seed") ) {
    case 0: // reset counts
      if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	pvminterface->BroadcastCommand("reset counts");
      }
      defaultsampler->ClearCounts();
      break;
    case 1: // reset seed
      if( (t = GetTokenVar(NULL, out,
			   "try reset seed 0")) ) {
	unsigned long int sigma = atol(t);
	if( sigma >= 0 ) {

	  if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	    ostringstream s;
	    s << "reset seed " << sigma;
	    pvminterface->BroadcastCommand(s.str().c_str());
	  }

	  defaultsampler->SetRandomSeed(sigma);

	} else {
	  out << "Token must be number > 0, try reset seed 0" << endl;
	}
      }
      break;
    }
    break;
  case 5: // set
    switch( GetToken(NULL, out, "try set sampler pagerank epsilon", 1,
		     "sampler") ) {
    case 0: // set sampler
      switch( GetToken(NULL, out, "try set sampler pagerank epsilon", 3,
		       "pagerank", "dateb", "tkleinberg") ) {
      case 0: // set sampler pagerank
	if( (t = GetTokenVar(NULL, out,
			     "try set sampler pagerank 0.3")) ) {
	  double epsilon = atof(t);
	  if( (epsilon > 0) && (epsilon < 1) ) {

	    if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	      ostringstream s;
	      s << "set sampler pagerank " << epsilon;
	      pvminterface->BroadcastCommand(s.str().c_str());
	    }

	    prsampler->SetParameters(epsilon);
	    defaultsampler = prsampler;

	  } else {
	    out << "Token must be 0 < epsilon < 1, try set sampler pagerank 0.3" << endl;
	  }
	}
	break;
      case 1: // set sampler dateb
	if( (t = GetTokenVar(NULL, out,
			     "try set sampler dateb 0.1 0.001")) ) {
	  double epsilon = atof(t);
	  if( (epsilon > 0) && (epsilon < 1) ) {
	    if( (t = GetTokenVar(NULL, out,
				 "try set sampler dateb 0.1 0.001")) ) {
	      double lambda = atof(t);
	      if( lambda > 0 ) {

		if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
		  ostringstream s;
		  s << "set sampler dateb " << epsilon << " " << lambda;
		  pvminterface->BroadcastCommand(s.str().c_str());
		}

		dbsampler->SetParameters(epsilon,lambda,gb->HighestDate() - gb->LowestDate());
		defaultsampler = dbsampler;

	      } else {
		out << "Token must be lambda > 0, try set sampler dateb 0.1 0.001" << endl;
	      }
	    }
	  } else {
	    out << "Token must be 0 < epsilon < 1, try set sampler pagerank 0.3" << endl;
	  }
	}
	break;
      case 2: // set sampler tkleinberg
	if( leaflist->size() > 0 ) {
	  out << "tkleinberg doesn't work with leafnodes/distributed computing" << endl;
	} else if( (t = GetTokenVar(NULL, out,
			     "try set sampler tkleinberg 0.1 auth")) ) {
	  double epsilon = atof(t);
	  if( (epsilon > 0) && (epsilon < 1) ) {
	    if( (t = GetTokenVar(NULL, out,
				 "try set sampler tkleinberg 0.1 auth")) ) {
	      ktype what = undef;
	      if( (strcmp(t,"auth") == 0) ) {
		what = auth;
	      } else if( (strcmp(t,"hubs") == 0) ) {
		what = hubs;
	      } else {
		out << "Token must be hubs or auth, try set sampler tkleinberg 0.1 auth" << endl;
	      }
	      if( what != undef ) {
		tksampler->SetParameters(epsilon,what);
		defaultsampler = tksampler;
	      }
	    }
	  } else {
	    out << "Token must be 0 < epsilon < 1, try set sampler pagerank 0.3" << endl;
	  }
	}
	break;
      }
      break;
    }
    break;
  case 6: // run
    switch( GetToken(NULL, out, "try sampler n", 2,
		     "sampler", "tagsampler") ) {
    case 0: // run sampler
      if( (t = GetTokenVar(NULL, out,
			   "try run sampler 5000")) ) {
	long num_samples = atol(t);
	if( num_samples > 0 ) {

	  // delegate samples to the slaves if there are any
	  if( (pvminterface->NumberOfTasks() > 0) && pvminterface->IsMaster() ) {
	    // compute total number of WebNodes
	    int totalgraphsize = web->size();
	    for(int i = 0; i < pvminterface->NumberOfOtherTasks(); i++ ) {
	      totalgraphsize += (pvminterface->StopID(i) - pvminterface->StartID(i));
	    }
	    // split number of samples into number of tasks uniformly 
	    // and send run command to each task
	    for(int i = 0; i < pvminterface->NumberOfOtherTasks(); i++) {
	      double p = static_cast<double>(pvminterface->StopID(i) - pvminterface->StartID(i))/totalgraphsize;
	      long howmany = gsl_ran_binomial(r, p, num_samples);
	      num_samples -= howmany;
	      ostringstream s;
	      s << "run sampler " << howmany;
	      pvminterface->SendCommand(i,s.str().c_str());
	    }
	    cerr << "pvm: finished delegating, ready to: run sampler " << num_samples << endl;
	  }
	  // do simulation on remaining samples
	  defaultsampler->SimulateForward(num_samples);
	  // delegate loose ends to others
	  PropagateLeafCounts();

	} else {
	  out << "Token must be 0 < n < " << LONG_MAX << ", try run sampler 5000" << endl;
	}
      }
      break;
    case 1: // run tagsampler
      if( (t = GetTokenVar(NULL, out,
			   "try run tagsampler 5000")) ) {
	long num_samples = atol(t);
	if( (num_samples > 0) && (leaflist->size() <= 0)) {
	  const uint32 *fromsetsizes = web->GetFromSetSizes();
	  if( fromsetsizes[0] > 0 ) {
	    defaultsampler->TaggedSimulateForward(num_samples, fromsetsizes);
	  } else {
	    out << "tagsampler cannot run. You must first tag some documents. See load tags command." << endl;
	  }
	} else {
	  out << "Token must be 0 < n < " << LONG_MAX << ", try run tagsampler 5000.\n"
	    "Note: cannot run tagsampler with leafnodes." << endl;
	}
      }
      break;
    }
    break;
  case 7: // load
    switch( GetToken(NULL, out, "try load [index|tags]", 2,
		     "index", "tags") ) {
    case 0: // load index
      if( (t = GetTokenVar(NULL, out, 
			   "try load index filename")) ) {
	ifstream g(t);
	if( !g.fail() ) {
	  if( !LoadLeaves(g) ) {
	    out << "Warning: index file was not properly loaded - leaf nodes may have inconsistent states" << endl;
	  }
	  g.close();
	} else {
	  out << "Error: could not open file [" << t << "]" << endl;
	}
      }
      break;
    case 1: // load tags
      if( (t = GetTokenVar(NULL, out, 
			   "try load tags filename")) ) {
	ifstream g(t);
	if( !g.fail() ) {
	  web->ClearTags(); 
	  if( !BuildTags(g) ) {
	    out << "Warning: some tagged documents not recognized/in dataset will be ignored" << endl;
	  }
	  g.close();
	  web->BuildFromSets(16);
	  web->MeasureFromSets();
	} else {
	  out << "Error: could not open file [" << t << "]" << endl;
	}
      }
      break;
    }
    break;
  case 8: // pvm
    switch( GetToken(NULL, out, "non-user command", 3,
		     "run", "save", "cat") ) {
    case 0: // pvm run
      switch( GetToken(NULL, out, "try sampler n", 1,
		       "sampler") ) {
      case 0: // pvm run sampler

	defaultsampler->ClearAllocForward();

	LeafCountsStruct* counts = new LeafCountsStruct[MAXLEAFMESSAGE_SIZE];
	uint32 size;

	// here we consolidate all leaf occupation messages we can get our hands on
	bool repeat = false;
	bool stop = false;
	do {

	  // read leaf occupation array
	  size = sizeof(LeafCountsStruct)*MAXLEAFMESSAGE_SIZE/sizeof(unsigned int);
	  pvminterface->GetLeafCounts((unsigned int*)counts, &size);
	  size = sizeof(unsigned int)*size/sizeof(LeafCountsStruct);
	  assert(size > 0);

	  // we use the fact that the leaf counts are sorted in decreasing
	  // ID number order in the counts array and in the web graph.
	  assert(counts[0].id <= gb->HighestID());
	  assert(counts[size-1].id >= gb->LowestID());

	  cerr << "pvm: loading " << size << " LeafNodes" << endl;
	  for(uint32 c = 0; c < size; c++) {
	    defaultsampler->IncrementAllocForward(counts[c].id - gb->LowestID(), 
						  counts[c].occupation_count);
	  }

	  commandline[0] = 0; // done processing command

	  // see if another pvm run sampler command was also sent
	  if( pvminterface->MessagePending() ) {
	    pvminterface->GetCommandWithTimeout(commandline);
	    assert(*commandline);
	    repeat = (strncmp(commandline,"pvm run sampler", 15) == 0);
	  } else {
	    stop = true;
	  }

	} while( repeat && !stop);
	delete[] counts;

	cerr << "pvm: now doing partial simulation" << endl;
	// do simulation on remaining samples
	defaultsampler->SimulateAllocForward();
	// delegate loose ends to others
	PropagateLeafCounts();

	// in case we intercepted another command by mistake, 
	// process it
	if( *commandline && !repeat ) {
	  return ProcessCommand(commandline,out);
	}
	break;
      }
      break;
    case 1: // pvm save
      switch( GetToken(NULL, out, "non-user command", 3,
		       "graph", "counts", "cat") ) {
      case 0: // pvm save graph
	if( (t = GetTokenVar(NULL, out, 
			   "try save graph filename")) ) {
	  ostringstream s;
	  s << tempdir << "ripper." << name << ".graph";
	  ofstream g(s.str().c_str());
	  if( !g.fail() ) {
	    web->PrintWebGraph(g);
	    g.close();
	    s.seekp(0);
	    s << "pvm cat ripper." << name << ".graph " << t;
	    pvminterface->SendCommand(pvminterface->LastMessageOriginator(), s.str().c_str());
	  } else {
	    cerr << "pvm: error: could not open " << s.str() << endl;
	  }
	} 
	break;
      case 1: // pvm save counts
	if( (t = GetTokenVar(NULL, out, 
			   "try save graph filename")) ) {
	  ostringstream s;
	  s << tempdir << "ripper." << name << ".counts";
	  ofstream g(s.str().c_str());
	  if( !g.fail() ) {
	    defaultsampler->PrintCounts(g);
	    g.close();
	    s.seekp(0);
	    s << "pvm cat ripper." << name << ".counts " << t;
	    pvminterface->SendCommand(pvminterface->LastMessageOriginator(), s.str().c_str());
	  } else {
	    cerr << "pvm: error: could not open " << s.str() << endl;
	  }
	} 
	break;
      }
      break;
    case 2: // pvm cat 
	if( (t = GetTokenVar(NULL, out, 
			   "try pvm cat source dest")) ) {
	  ostringstream s;
	  s << "cat " << tempdir << t;
	  if( (t = GetTokenVar(NULL, out, 
			   "try pvm cat source dest")) ) {
	    s << " >> " << t;
	    cerr << "pvm: " << s.str() << endl;
	    if( system(s.str().c_str()) ) {
	      out << "Couldn't execute command: " << s.str() << endl;
	    }
	  }
	}
      break;
    }
    break;
  }

  commandline[0] = 0; // indicate command is done
  return quit;
}


/// prints out some memory usage statistics
void Talker::PrintStatistics(ostream& o) {
  o << "Number of documents processed:           " << num_docs_processed << endl;
  gb->StatisticsMem(o);
  o << "\nTotal number of leafnodes added:         " << leaflist->size() << endl;
  o << "Total heap needed for all leafnodes:     " << sizeof(LeafNode)*leaflist->size() << endl;
}

/// prints out some link graph statistics
void Talker::PrintStatisticsGraph(ostream& o) {
  gb->StatisticsGraph(o);
}


/// reads a file containing space separated
/// urls and tags those which are found in 
/// the web link graph
bool Talker::BuildTags(ifstream& g) {
  string s;
  bool problems = false;
  while( g && !g.eof() ) {
    g >> s;
    if( s[0] == '#' ) { // comment ignore
      getline(g,s); 
      continue; 
    } 
    WebNodePtr w = gb->FindWebNode(s.c_str());
    if( w ) {
      w->SetTag(0);
    } else {
      problems = true;
      cerr << "warning: unrecognized tag [" << s << "]" <<  endl;
    }
  }
  return !problems;
}


/// In interactive mode, reads the next token from the 
/// command buffer, and returns the token's id number if 
/// recognized.
int Talker::GetToken(char* command, ostream& out, char* message, 
		int n, ...) {
  char *t = strtok(command, " ");
  if( !t ) {
    out << "Missing token, " << message << endl;
  } else { 
    va_list ap;
    va_start(ap, n);
    for(int k = 0; k < n; k++) {
      if( strcmp(t, va_arg(ap, char*)) == 0 ) {
	va_end(ap);
	return k;
      } 
    }
    va_end(ap);
    out << "Unrecognized token " << t << ", " << message << endl;
  }
  return -1;
}

/// In interactive mode, reads the next token from the
/// command buffer, and returns the token's string value for
/// processing.
char *Talker::GetTokenVar(char* command, ostream& out, char* message) {
  char *t = strtok(command, " ");
  if( !t ) {
    out << "Missing token, " << message << endl;
  } else if( strlen(t) > 0 ) {
    return t;
  } else {
    out << "Missing token, " << message << endl;
  }
  return NULL;
}

typedef struct IDPair {
  uint32 one;
  uint32 two;
};

#define EQUIV_SIZE 1024

bool Talker  :: LoadLeaves(ifstream& g) {
  string alias,url;
  uint32 nodeID;
  uint16 date;
  bool problems = false;

  gb->SetupLeafTable();
  
  // duplicate nodes are saved here
  IDPair *equivalences = new IDPair[EQUIV_SIZE];
  int equiv_pointer = 0;
  ofstream neq(nodequiv.str().c_str(),std::ios::app);

  while( g && !g.eof() ) {

    g >> url; 
    if( url[0] == '#' ) { // comment ignore
      getline(g,url); 
      continue; 
    } 
    g >> nodeID >> date;
    getline(g,alias); // optional

    WebNodePtr w = gb->FindWebNode(url.c_str());
    if( w ) {
      // after url redirects are taken into account, the same url
      // can exist as several distinct nodes - this is not an error
      equivalences[equiv_pointer].one = w->ID();
      equivalences[equiv_pointer].two = nodeID;
      if( ++equiv_pointer >= EQUIV_SIZE ) {
	for(int z = 0; z < EQUIV_SIZE; z++) {
	  neq << equivalences[z].one << " = " << equivalences[z].two << endl;
	}
	equiv_pointer = 0;
      }
    } else {
      ptrdiff_t key = gb->FindLeafNodeKey(url.c_str());
      if( key >= 0 ) { // found url in trie, insert a new leaf
	LeafNodePtr leaf = new LeafNode(nodeID);
	leaf->SetDate(date);
	gb->AddLeaf(key, leaf);
	leaflist->push_front(leaf);
      } else { 
	// url not found in trie
	// ignore leaf
      }
    }
  }

  // write the duplicate nodes
  for(int z = 0; z < equiv_pointer; z++) {
    neq << equivalences[z].one << " = " << equivalences[z].two << endl;
  }
  delete[] equivalences;
  neq.close();

  gb->UpdateLeafLinks();
  return !problems;
}

