/** @file trie.cc 
 * This file contains the member function definitions
 * for the class Trie.
 */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "trie.h"
#include <cstring>
#include <cassert>

Trie :: Trie(unsigned long slen, long jels) {
  // allocate bigstring
  slen_ = slen;
  bigs = new char[slen];
  memset(bigs, 0, slen);
  end_of_bigs = 0;
  // allocate the hashes
  jumptable = new SimpleCharPtrHashTable(jels);

  // clear some statistics
  stats_bigstring_insertions = 0;
  stats_jumptable_insertions = 0;
  stats_cumulative_string_size = 0;
}

void Trie :: Statistics(ostream& o) {
  o << "Total number of URL strings inserted:    " << stats_bigstring_insertions << endl;
  o << "Total original size of URLs inserted:    " << stats_cumulative_string_size << endl;
  o << "Heap needed for bigstring (bytes):       " << end_of_bigs << endl;
  o << "Heap reserved for bigstring (bytes):     " << slen_ << endl;

  o << "\nTotal number of insertions in jumptable: " << stats_jumptable_insertions << endl;
  o << "Heap needed for jumptable (bytes):       " << (stats_jumptable_insertions * sizeof(SimpleHashPair<char*>)) << endl;
  o << "Heap reserved for jumptable (bytes):     " << (jumptable->Size() * sizeof(SimpleHashPair<char*>)) << endl;
}


/// Searches for a URL within the trie. 
/**
 * If found, returns a character pointer to the end of the URL string
 * within the trie, else returns -1.
 */
// code lifted from InsertURL, someday this will have to be refactored
ptrdiff_t Trie :: FindURL(const char *url) {
  char *p = bigs;
  const char *q = url;

  //  url = CompressURL(url);

  while( *q ) {
    // skip over common prefix
    while( (*q != 0) && ( *p == *q ) ) {p++; q++; };

    // either q string is empty, or p string is empty, or both

    if( (*q == 0) && (*p == 0) ) { 

      return p - bigs;

    } else if( *q == 0 ) { // a longer p string was inserted first
      // follow as many branches as possible

      char *nextp = jumptable->Find(p - bigs);

      while( nextp && (*nextp != 0) ) {
	p = nextp;

	nextp = jumptable->Find(p - bigs);

      }

      if( nextp && (*nextp  == 0) ) { // found the string

	return p - bigs;

      } else { // there are no further branches at p, 

	return -1; // string not found

      }	
    } else { // 0 != *q != *p

      char *nextp = jumptable->Find(p - bigs);

      if( nextp ) { // branch exists
	p = nextp;
	// and back to while
      } else { // no branch at p

	return -1; // string not found

      }
    }
  }
  
  return -1; // should not get to here
}

// char & ~SLASHBIT unsets branch bit
// char | SLASHBIT sets branch bit 
// char & SLASHBIT is true only if branch bit is set

/// Inserts a URL string into the trie. 
/**
 * Always returns a char pointer to the end of the inserted string.
 * The second argument is of urltype, and signifies that the 
 * URL that is being inserted is either the current 
 * document's url, or a url found in one of the 
 * current document's anchor tags.
 *
 * Note InsertURL assumes url is a pure ASCII string (7 bits).
 */
ptrdiff_t Trie :: InsertURL(const char *url) throw (overflow_error) {
  char *p = bigs;
  const char *q = url;
  char *end_of_inserted_string = NULL;

  stats_cumulative_string_size += (strlen(url)+1);

  // DEBUG cout << (node ? "========== " : "- - - - - -") << url << endl;

  if( end_of_bigs + strlen(url) > slen_ ) {
    Statistics(cout);
    cerr << "d'oh: out of string memory" << endl;
    throw overflow_error("");
  } else if ( !end_of_bigs ) { // first one is special
    // just append string
    // DEBUG cout << "appending " << q << endl;
    while( *q ) { bigs[end_of_bigs++] = *q++; }
    bigs[end_of_bigs] = 0;

    end_of_inserted_string = bigs + end_of_bigs; 

    end_of_bigs++;
    stats_bigstring_insertions++;

  } else {

    while( *q ) {
      // skip over common prefix
      // try      while( (*q != 0) && ( (*p & ~NODEBIT) == *q ) ) {p++; q++; };
      while( (*q != 0) && ( *p == *q ) ) {p++; q++; };

      // DEBUG cout << "skipped: q=[" << q << "] vs p=[" << p << "]" << endl;

      // either q string is empty, or p string is empty, or both

      if( (*q == 0) && (*p == 0) ) { 

	// we found the string
	// DEGUG cout << "found" << endl;

	end_of_inserted_string = p; 

      } else if( *q == 0 ) { // a longer p string was inserted first
	// follow as many branches as possible
	// DEBUG cout << "checking for a branch at p=[" << p << endl;

	char *nextp = jumptable->Find(p - bigs);

	while( nextp && (*nextp != 0) ) {
	  p = nextp;

	  nextp = jumptable->Find(p - bigs);

	  // DEBUG cout << "branching p=[" << p << "]->[" << nextp << "]" << endl;
	}

	if( nextp && (*nextp  == 0) ) { // found the string
	  // DEBUG cout << "found" << endl;

 	  p = nextp; 
	  end_of_inserted_string = nextp; 

	} else { // there are no further branches at p, append q's terminating null 
	  assert( !nextp );
	  // DEBUG cout << "appending [" << q << "]" << endl;

	  jumptable->Insert(p - bigs, bigs + end_of_bigs);
	  
	  stats_jumptable_insertions++;
	  bigs[end_of_bigs] = 0;

	  end_of_inserted_string = bigs + end_of_bigs; 

	  end_of_bigs++;
	  stats_bigstring_insertions++;
	}	
      } else { // 0 != *q != *p

	char *nextp = jumptable->Find(p - bigs);

	// DEBUG cout << "branching p=[" << p << "]->[" << nextp << "]" << endl;
	if( nextp ) { // branch exists
	  p = nextp;
	  // and back to while
	} else { // no branch at p, so we append q
	  // DEBUG cout << "appending [" << q << "]" << endl;

	  jumptable->Insert(p - bigs, bigs + end_of_bigs);

	  stats_jumptable_insertions++;
	  while( *q ) { bigs[end_of_bigs++] = *q++; }
	  bigs[end_of_bigs] = 0;

	  end_of_inserted_string = bigs + end_of_bigs; 

	  end_of_bigs++;
	  stats_bigstring_insertions++;
	  // *q == 0 now
	}
      }
    }

  }

  assert(end_of_inserted_string);
  assert(*end_of_inserted_string == 0);

  return end_of_inserted_string - bigs;

}



#ifdef UNIT_TEST

#define MAKE_TEST_STATEMENT(x) \
do { \
  cout << "testing: " #x << endl; \
  cout << "output: " << endl; \
  (x); \
} while(false)

#define MAKE_TEST(x,y) \
do { \
   cout << argv[0] << ": " << (((x) == (y)) ? "PASSED" : "FAILED") << " " #x <<endl; \
   cout << "output: " << (x) << endl; \
} while(false)

#define MAKE_TEST_EX(x,y) \
do { \
   try { \
     (x); \
     cout << argv[0] << ": FAILED" << " " #x << endl; \
   } catch (y) { \
     cout << argv[0] << ": PASSED" << " " #x << endl; \
   } \
} while(false)

int main(int argc, char** argv) {
  Trie *trie = new Trie(100, 5);
  
  MAKE_TEST(trie->InsertURL("banana"),6);
  MAKE_TEST(trie->InsertURL("apple"),12);
  MAKE_TEST(trie->InsertURL("bananarama"),17);
  
  MAKE_TEST(trie->FindURL("apple"),12);
  MAKE_TEST(trie->FindURL("coconut"),-1);

  MAKE_TEST(trie->StatsCumulativeStringSize(), 48);
  MAKE_TEST(trie->StatsBigstringInsertions(), 3);
  MAKE_TEST(trie->StatsJumptableInsertions(), 2);
  
  MAKE_TEST_EX(trie->InsertURL("abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz"),overflow_error);
}
#endif
