/** @file trie.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifndef _TRIE_H_
#define _TRIE_H_
#include <iostream>
#include "simplehash.h"
#include <stdexcept>

#define STRINGBUF_LEN0 511
#define STRINGBUF_LEN1 1023
#define STRINGBUF_LEN2 2047
#define STRINGBUF_LEN3 4095

/// Hashtable used to navigate trie.
typedef SimpleHashTable<char*> SimpleCharPtrHashTable;

/// Stores URL strings by superposition of common prefixes
/**
 * This class implements a classic trie structure (Knuth, Vol. 3)
 * which consists of a very long string space together with a hashtable
 * (jumptable) which allows navigation.
 *
 * The trie allows large space savings by storing common string prefixes only
 * once.
 */
class Trie {
 public:
  Trie(unsigned long slen, long jlen);
  ptrdiff_t FindURL(const char *url);
  ptrdiff_t InsertURL(const char *url) throw (overflow_error);

  void Statistics(ostream& o);

  uint32 StatsCumulativeStringSize() 
    { return stats_cumulative_string_size; }
  uint32 StatsBigstringInsertions()
    { return stats_bigstring_insertions; }
  uint32 StatsJumptableInsertions()
    { return stats_jumptable_insertions; }

  char* bigs;

 protected:

  unsigned long slen_;
  unsigned long end_of_bigs;

  SimpleCharPtrHashTable *jumptable;

  uint32 stats_cumulative_string_size;
  uint32 stats_bigstring_insertions;
  uint32 stats_jumptable_insertions;

};
#endif
