/** @file webnode.h */
/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */
#ifndef _WEB_NODE_H_
#define _WEB_NODE_H_
#include "mempool.h"
#include "simplehash.h"
#include "leafnode.h"
#include <stdexcept>
#include <set>

/// Encapsulates a web document.
class WebNode; // defined below
/// All WebNodes are allocated on the heap as WebNodePtrs.
typedef WebNode* WebNodePtr;

/// A pointer to a web document/URL.
/**
 * During the construction of the web graph,
 * document links are of type ptrdiff_t, but after
 * the WebLinkGraph is undocked, they are of type WebNodePtr
 */
struct LinkStruct {
  union {
    ptrdiff_t pointer_diff;
    WebNodePtr webnode_ptr;
    LeafNodePtr leafnode_ptr;
  };
};

/// A memory management wrapper around LinkStruct.
typedef MemPoolObject<LinkStruct> Link;

/// A scratch structure used by samplers.
/// Each sampler should cast it into something useful
typedef uint32 ScratchStruct;

#define Mb 1048576L
#define WEBNODE_MEMPOOL_DELTA ((1 * Mb)/sizeof(WebNode))
#define LINK_MEMPOOL_DELTA ((1 * Mb)/sizeof(Link))

#define TAG_NUMBER_OF_BITS 16

/// Contains all the information about a web document.
struct WebNodeStruct {
  // the ordering here is important due to alignment issues
  Link *tolinks;
  Link *fromlinks;
  uint32 id;
  uint32 occupation_count;
  uint16 date;
  uint16 num_tolinks;
  uint16 num_fromlinks;
  uint16 num_valid_tolinks;
  uint16 num_leaflinks;
  uint16 tag;
  ScratchStruct scratch;
};

/// Used by GraphBuilder to store uniquely the anchor links
/// and then insert them into the WebNode.
typedef set<ptrdiff_t> RawLinkSet; //,CharPtrComparisonFunction> RawLinkSet;

/// Encapsulates a web document.
/**
 * Every web document read by the ripper is represented by
 * a WebNode. The construction of a WebNode is complicated,
 * and is done by GraphBuilder, which also links the nodes
 * into a WebLinkGraph.
 * All the data members are defined as a WebNodeStruct, WebNode
 * is really just a wrapper for WebNodeStruct to handle custom
 * memory management.
 * The class inherits memory management from MemoryPooled<T>.
 */
class WebNode: public MemoryPooled<WebNodeStruct> {
 public:

  WebNode(uint32 idno);
  void InsertRawLinks(RawLinkSet *s);
  void NormalizeRawLinks(SimpleHashTable<WebNodePtr> *h);
  size_t RealSize();
  // inline functions
  int NumberOfValidToLinks() 
    { return data.num_valid_tolinks;}
  int NumberOfDanglingToLinks() 
    { return (data.num_tolinks - data.num_valid_tolinks - data.num_leaflinks);}
  int NumberOfValidFromLinks() 
    { return data.num_fromlinks;}
  int NumberOfLeafLinks() 
    { return data.num_leaflinks;}

  void IncrementNumberOfFromLinks() 
    { data.num_fromlinks++; }

  void AppendFromLink(WebNodePtr anothernode) throw (overflow_error);

  void UpdateLeafLinks(SimpleLeafNodePtrHashTable *leaftable);

  void SetDate(uint16 adate);

  WebNodePtr ValidToLink(int k) 
    {
      assert(k < data.num_valid_tolinks);
      assert(data.tolinks[k].data.webnode_ptr);
      return data.tolinks[k].data.webnode_ptr;
    }
  WebNodePtr ValidFromLink(int k) 
    {
      assert(data.fromlinks[k].data.webnode_ptr);
      return data.fromlinks[k].data.webnode_ptr;
    }
  LeafNodePtr ValidLeafLink(int k)
    {
      assert(k < data.num_leaflinks);
      assert(data.tolinks[data.num_valid_tolinks + k].data.leafnode_ptr);
      return data.tolinks[data.num_valid_tolinks + k].data.leafnode_ptr;
    }
  LeafNodePtr ValidLeafLinkDirectly(int k) ///< Same as ValidLeafLink but saves a +/- in -O3
    {
      assert(k >= data.num_valid_tolinks);
      assert(k < data.num_valid_tolinks + data.num_leaflinks);
      assert(data.tolinks[k].data.leafnode_ptr);
      return data.tolinks[k].data.leafnode_ptr;
    }

  uint32 ID() 
    { return data.id; }
  uint16 Date() 
    { return data.date; }

  void ClearTag() 
    { data.tag = 0; }
  void SetTag(int k) 
    {
      assert(k < TAG_NUMBER_OF_BITS);
      data.tag |= (1<<k); // sets the kth bit
    }
  bool Tagged(int k)
    {
      return ((data.tag & (1<<k)) != 0);
    }

  void ClearOccupationCount() 
    { 
      data.occupation_count = 0;
      data.scratch = 0;
    }
  uint32 OccupationCount() 
    { return data.occupation_count; }
  void IncrementOccupationCount() 
    { data.occupation_count++; }
  void IncrementOccupationCount(int c) 
    { data.occupation_count += c; }

  ScratchStruct Scratch() 
    { return data.scratch; }
  void SetScratch(ScratchStruct ascratch) 
    { data.scratch = ascratch; }

 private:
  static MemPool<LinkStruct> global_link_pool;
};

#endif
